# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=too-many-lines

"""
Generic Classes for reading XPS files into python dictionary.
"""

from typing import List, Dict

from pynxtools_xps.sle.sle_specs import SleMapperSpecs
from pynxtools_xps.txt.txt_scienta import TxtMapperScienta

# from pynxtools_xps.txt.txt_specs import TxtMapperSpecs
from pynxtools_xps.txt.txt_vamas_export import TxtMapperVamasExport
from pynxtools_xps.vms.vamas import VamasMapper
from pynxtools_xps.xy.xy_specs import XyMapperSpecs
from pynxtools_xps.xml.xml_specs import XmlMapperSpecs


class XpsDataFileParser:
    """Class intended for receiving any type of XPS data file."""

    __prmt_file_ext__ = ["sle", "txt", "vms", "xml", "xy"]
    __prmt_metadata_file_ext__ = ["slh"]
    __vendors__ = ["kratos", "scienta", "specs", "unkwown"]
    __prmt_vndr_cls: Dict[str, Dict] = {
        "sle": {"specs": SleMapperSpecs},
        "txt": {
            "scienta": TxtMapperScienta,
            # 'specs': TxtMapperSpecs,
            "unknown": TxtMapperVamasExport,
        },
        "vms": {"unkwown": VamasMapper},
        "xml": {"specs": XmlMapperSpecs},
        "xy": {"specs": XyMapperSpecs},
    }

    __file_err_msg__ = (
        "Need an XPS data file with the following extension: " f"{__prmt_file_ext__}"
    )

    __vndr_err_msg__ = (
        "Need an XPSdata file from the following vendors: " f"{__vendors__}"
    )

    def __init__(self, file_paths: List) -> None:
        """
        Receive XPS file path.

        Parameters
        ----------
        file_paths : List
            XPS file path.
        """
        if isinstance(file_paths, str):
            file_paths = [file_paths]

        self.files = file_paths
        self.config_file = None

        if not self.files:
            raise ValueError(XpsDataFileParser.__file_err_msg__)

    def get_dict(self, **kwargs) -> dict:
        """
            Return python dict fully filled data from xps file.
        Returns
        -------
        python dictionary
        """
        for file in self.files:
            file_ext = file.rsplit(".")[-1]
            if file_ext in XpsDataFileParser.__prmt_file_ext__:
                vendor = XpsDataFileParser.check_for_vendors(file)
                try:
                    parser_class = XpsDataFileParser.__prmt_vndr_cls[file_ext][vendor]
                    parser_obj = parser_class()
                    parser_obj.parse_file(file, **kwargs)
                    self.config_file = parser_obj.config_file
                    return parser_obj.data_dict

                except ValueError as val_err:
                    raise ValueError(XpsDataFileParser.__vndr_err_msg__) from val_err
                except KeyError as key_err:
                    raise KeyError(XpsDataFileParser.__vndr_err_msg__) from key_err
            else:
                raise ValueError(XpsDataFileParser.__file_err_msg__)
        return {}

    @classmethod
    def check_for_vendors(cls, file: str) -> str:
        """
        Check for the vendor name of the XPS data file.

        """
        file_ext = file.rsplit(".")[-1]

        vendor_dict = XpsDataFileParser.__prmt_vndr_cls[file_ext]

        if len(vendor_dict) == 1:
            return list(vendor_dict.keys())[0]
        if file_ext == "txt":
            return cls._check_for_vendors_txt(file)
        return None

    @classmethod
    def _check_for_vendors_txt(cls, file: str) -> str:
        """
        Search for a vendor names in a txt file

        Parameters
        ----------
        file : str
            XPS txt file.

        Returns
        -------
        vendor
            Vendor name if that name is in the txt file.

        """
        vendor_dict = XpsDataFileParser.__prmt_vndr_cls["txt"]

        with open(file, encoding="utf-8") as txt_file:
            contents = txt_file.read()

        for vendor in vendor_dict:
            vendor_options = [vendor, vendor.upper(), vendor.capitalize()]

            if any(vendor_opt in contents for vendor_opt in vendor_options):
                return vendor
        return "unknown"



"""
Basic example based test for the XPS reader
"""
import os
import xml.etree.ElementTree as ET
from glob import glob

from pynxtools.dataconverter.helpers import (
    generate_template_from_nxdl,
    validate_data_dict,
)
from pynxtools.dataconverter.template import Template
from pynxtools.nexus.nxdl_utils import get_nexus_definitions_path
from pynxtools_xps.reader import XPSReader


def test_example_data():
    """
    Test the example data for the XPS reader
    """
    reader = XPSReader
    assert callable(reader.read)

    def_dir = get_nexus_definitions_path()

    data_dir = os.path.join(os.path.dirname(__file__), "data")

    reader_dirs = sorted(glob(os.path.join(data_dir, "*")))

    for reader_dir in reader_dirs:
        input_files = sorted(glob(os.path.join(reader_dir, "*")))

        for supported_nxdl in reader.supported_nxdls:
            nxdl_file = os.path.join(
                def_dir, "contributed_definitions", f"{supported_nxdl}.nxdl.xml"
            )

            root = ET.parse(nxdl_file).getroot()
            template = Template()
            generate_template_from_nxdl(root, template)

            read_data = reader().read(
                template=Template(template), file_paths=tuple(input_files)
            )

            assert isinstance(read_data, Template)
            assert validate_data_dict(template, read_data, root)


## This will be implemented in the future.
# =============================================================================
# def test_vms_mapper():
#     mapper = VamasMapper
#     data_dir = os.path.join(os.path.dirname(__file__), "data", "vms")
#
#     files_and_keys = {
#         "regular": {
#             "/ENTRY[entry]/Group_1 as-loaded/regions/Region_Survey/scan_mode": "REGULAR",
#             "data['2 S1110, UHV, RT, Epass = 20 eV__MgKLL_1']['cycle0']": 1351,
#         },
#         "irregular": {
#             "/ENTRY[entry]/Group_1 as-loaded/regions/Region_Survey/scan_mode": "IRREGULAR",
#             "/ENTRY[entry]/Group_1 as-loaded/regions/Region_Fe2p/instrument/analyser/energydispersion/scan_mode": "FixedAnalyserTransmission",
#             "data['2 S1110, UHV, RT, Epass = 20 eV__Fe3s-Si2p-Mg2s']['cycle0']": 761,
#         },
#     }
#
#     # (C["benz"], "20240122_SBenz_102_20240122_SBenz_SnO2_10nm.vms"),
#     # (r"C:\Users\pielsticker\Lukas\MPI-CEC\Projects\deepxps\RUB MDI\example_spectra_Florian\Co", "Co 2p 0008751 M1.vms"),
#     # (C["pielst"], C["EX889"], "vms", f"{C['EX889']}_regular.vms"),
#     ((C["pielst"], C["EX889"], "vms", f"{C['EX889']}_irregular.vms"),)
#     # (C["schu"], "CleanData-alphaII VOPO4 C2 BC4316.vms"),
#     # (C["pielst"], C["EX889"], "vms", "d_reg.vms"),
#     ((C["pielst"], C["EX889"], "vms", "d_irreg.vms"),)
#
#     for vms_file in os.listdir(data_dir):
#         data = mapper.parse_file(file=file)
#
#
# for k, v in d.items():
#     if isinstance(v, str):
#         # print(k)
#         if "REG" in v:
#             print(k)
# =============================================================================



"""
Data model for CasaXPS.
"""
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=too-many-instance-attributes

import re
from typing import List, Dict, Any
from dataclasses import dataclass, field

from pynxtools_xps.reader_utils import XpsDataclass


class CasaProcess:
    """Processing and fitting information from CasaXPS."""

    def __init__(self):
        self.n_alignments: int = 0
        self.n_unknown_processes: int = 0
        self.n_regions: int = 0
        self.n_components: int = 0

        self.alignments: List[Any] = []
        self.unknown_processes: List[Any] = []
        self.regions: List[Any] = []
        self.components: List[Any] = []

    def process_comments(self, comment_list: List[str]):
        line_map = {
            "Calib": self.process_alignment,
            "CASA region": self.process_region,
            "CASA comp": self.process_component,
        }

        for line in comment_list:
            for token, func in line_map.items():
                if line.startswith(token):
                    func(line)

        casa_data = {
            "alignments": self.alignments,
            "regions": self.regions,
            "components": self.components,
        }

        self.count_occurences()

        return casa_data

    def count_occurences(self):
        """Count occurence of different process items."""
        self.n_alignments = len(self.alignments)
        self.n_unknown_process = len(self.unknown_processes)
        self.n_regions = len(self.regions)
        self.n_components = len(self.components)

    def process_alignment(self, align_str: str):
        """
        Process one alignment.

        Parameters
        ----------
        align_str : str
            String with alignment definition in a VMS file.
            Typical example:
                'Calib M = 6.8 A = 0 BE ADD'


        Returns
        -------
        None.

        """
        split_line = self._split_line(align_str)

        align = CasaAlignment()

        measured_index = split_line.index("M")
        aligned_index = split_line.index("A")
        align.measured_energy = float(split_line[measured_index + 2])
        align.aligned_energy = float(split_line[aligned_index + 2])

        align.energy_type = map_energy_type(split_line[7])
        align.operation = split_line[8]

        if align.operation == "ADD":
            align.energy_offset = align.measured_energy - align.aligned_energy
            align.operation = "addition"

        align.validate_types()

        self.alignments += [align]

    def process_region(self, region_str: str):
        """
        Process one fit region.

        Parameters
        ----------
        region_str : str
            String with region definition in a VMS file.
            Typical example:
                'CASA region (*O1s*) (*Shirley*) 718.88221 726.6 9.74 2 0 9 1450.5943 -450 0 0 (*Mo 3d*) 95.9219 0 9.74'

        Returns
        -------
        None.

        """
        split_line = self._split_line(region_str)

        region = CasaRegion()
        region.name = split_line[2]
        region.bg_type = split_line[3]
        region.start = float(split_line[4])
        region.end = float(split_line[5])
        region.rsf = float(split_line[6])
        region.av_width = float(split_line[7])
        region.start_offset = float(split_line[8])
        region.end_offset = float(split_line[9])
        region.cross_section = split_line[10:14]
        region.tag = split_line[14]
        region.unknown_0 = float(split_line[15])
        region.unknown_1 = float(split_line[16])
        region.rsf_effective = float(split_line[17])

        region.validate_types()

        self.regions += [region]

    def process_component(self, component_str: str):
        """
        Process one fit component if it is defined in the

        Parameters
        ----------
        component_str : str
            String with component definition in a VMS file.
            Typical example:
                'CASA comp (*O1s, lattice*) (*GL(50)*) Area 308996.39 1e-020 15401311 -1 1 MFWHM 1.5315216 0.32 2 -1 1 Position 723.30224 719.3797 726.3 -1 1 RSF 0.063 MASS 15.9994 INDEX -1 (*O 1s*) CONST (**) UNCORRECTEDRSF 2.85'


        Returns
        -------
        None.

        """
        split_line = self._split_line(component_str)

        component = CasaComponent()
        component.name = split_line[2]
        component.lineshape = split_line[3]

        area_index = split_line.index("Area")
        component.area = float(split_line[area_index + 1])
        component.area_min = float(split_line[area_index + 2])
        component.area_max = float(split_line[area_index + 3])
        component.area_ref_comp_id = int(split_line[area_index + 4])
        component.area_ref_comp_factor = float(split_line[area_index + 4])

        width_index = split_line.index("MFWHM")
        component.width = float(split_line[width_index + 1])
        component.width_min = float(split_line[width_index + 2])
        component.width_max = float(split_line[width_index + 3])
        component.width_ref_comp_id = int(split_line[width_index + 4])
        component.width_ref_comp_factor = float(split_line[width_index + 5])

        position_index = split_line.index("Position")
        component.position = float(split_line[position_index + 1])
        component.position_min = float(split_line[position_index + 2])
        component.position_max = float(split_line[position_index + 3])
        component.position_ref_comp_id = int(split_line[position_index + 4])
        component.position_ref_comp_factor = float(split_line[position_index + 5])

        rsf_index = split_line.index("RSF")
        component.rsf = float(split_line[rsf_index + 1])
        mass_index = split_line.index("MASS")
        component.mass = float(split_line[mass_index + 1])
        comp_index = split_line.index("INDEX")
        component.index = int(split_line[comp_index + 1])
        component.tag = str(split_line[comp_index + 2])
        const_index = split_line.index("CONST")
        component.const = str(split_line[const_index + 1])
        uncorrected_rsf_index = split_line.index("UNCORRECTEDRSF")
        component.uncorrected_rsf = float(split_line[uncorrected_rsf_index + 1])

        component.validate_types()
        self.components += [component]

    def _split_line(self, line):
        """Split line in Casa processing."""
        regex_pattern = r"\s+|\(\*(?:(?!\*\)).)*?\*\)|\S+"
        result = [match for match in re.findall(regex_pattern, line) if match.strip()]
        return [re.sub(r"[\(*?\*)]", "", text) for text in result]


def map_energy_type(energy_str):
    """Map energy names to NXDL-compliant concepts."""
    replacements = {
        "BE": "binding",
        "KE": "kinetic",
        "binding energy": "binding",
        "binding": "kinetic",
        "Binding": "binding",
        "Kinetic": "kinetic",
    }
    return replacements.get(energy_str)


@dataclass
class CasaAlignment(XpsDataclass):
    """An object to store one CasaXPS alignment."""

    energy_type: str = "binding"
    energy_offset: float = 0.0
    energy_offset_units: str = "eV"
    measured_energy: float = 0.0
    measured_energy_units: str = "eV"
    aligned_energy: float = 0.0
    aligned_energy_units: str = "eV"
    operation: str = "eV"


@dataclass
class CasaRegion(XpsDataclass):
    """An object to store one CasaXPS peak fit region."""

    name: str = ""
    rsf: float = 0.0
    start: float = 0.0
    start_units: str = "eV"
    end: float = 0.0
    end_units: str = "eV"
    bg_type: str = ""
    av_width: float = 0.0
    av_width_units: str = "eV"
    start_offset: float = 0.0
    start_offset_units: str = "CPS"
    end_offset: float = 0.0
    end_offset_units: str = "CPS"
    cross_section: list = field(default_factory=list)
    tag: str = ""
    unknown_0: float = 0.0
    unknown_1: float = 0.0
    rsf_effective: float = 0.0


@dataclass
class CasaComponent(XpsDataclass):
    """An object to store one CasaXPS peak fit component."""

    name: str = ""
    index: int = -1
    lineshape: str = ""

    area: float = 0.0
    area_min: float = 0.0
    area_max: float = 0.0
    area_ref_comp_id: int = -1
    area_ref_comp_factor: float = 0.0

    width: float = 0.0
    width_units: str = "eV"
    width_min: float = 0.0
    width_min_units: str = "eV"
    width_max: float = 0.0
    width_max_units: str = "eV"
    width_ref_comp_id: int = -1
    width_ref_comp_factor: float = 0.0

    position: float = 0.0
    position_units: str = "eV"
    position_min: float = 0.0
    position_min_units: str = "eV"
    position_max: float = 0.0
    position_max_units: str = "eV"
    position_ref_comp_id: int = -1
    position_ref_comp_factor: float = 0.0

    rsf: float = 0.0
    uncorrected_rsf: float = 0.0
    mass: float = 0.0
    tag: str = ""
    const: str = ""  # CONST



"""
Helper functions for populating NXmpes template
"""
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union
from pathlib import Path
from scipy.interpolate import interp1d
import numpy as np

from dataclasses import dataclass


@dataclass
class XpsDataclass:
    """Generic class to hold a data model and a type validation method."""

    def validate_types(self):
        ret = True
        for field_name, field_def in self.__dataclass_fields__.items():
            actual_type = type(getattr(self, field_name))
            if actual_type != field_def.type:
                print(f"\t{field_name}: '{actual_type}' instead of '{field_def.type}'")
                ret = False
        return ret

    def __post_init__(self):
        if not self.validate_types():
            raise ValueError("Wrong types")

    def dict(self):
        return self.__dict__.copy()


class XPSMapper(ABC):
    """Abstract base class from mapping from a parser to NXmpes template"""

    def __init__(self):
        self.file: Union[str, Path] = ""
        self.raw_data: List[str] = []
        self._xps_dict: Dict[str, Any] = {}
        self._root_path = "/ENTRY[entry]"

        self.parser = None

    @abstractmethod
    def _select_parser(self):
        """
        Select the correct parser for the file extension and format.

        Should be implemented by the inheriting mapper.

        Returns
        -------
        Parser

        """

    @property
    def data_dict(self) -> dict:
        """Getter property."""
        return self._xps_dict

    def parse_file(self, file, **kwargs):
        """
        Parse the file using the Scienta TXT parser.

        """
        self.file = file
        self.parser = self._select_parser()
        self.raw_data = self.parser.parse_file(file, **kwargs)

        file_key = f"{self._root_path}/File"
        self._xps_dict[file_key] = file

        self.construct_data()

        return self.data_dict

    @abstractmethod
    def construct_data(self):
        """
        Map from individual parser format to NXmpes-ready dict.

        Should be implemented by the inheriting mapper.

        """


def to_snake_case(str_value):
    """Convert a string to snake_case."""
    if " " in str_value:
        return "_".join(word.lower() for word in str_value.split())
    return convert_pascal_to_snake(str_value)


def convert_pascal_to_snake(str_value):
    """Convert pascal case text to snake case."""
    pattern = re.compile(r"(?<!^)(?=[A-Z])")
    return pattern.sub("_", str_value).lower()


def convert_snake_to_pascal(str_value):
    """Convert snakecase text to pascal case."""
    return str_value.replace("_", " ").title().replace(" ", "")


def safe_arange_with_edges(start, stop, step):
    """
    In order to avoid float point errors in the division by step.

    Parameters
    ----------
    start : float
        Smallest value.
    stop : float
        Biggest value.
    step : float
        Step size between points.

    Returns
    -------
    ndarray
        1D array with values in the interval (start, stop),
        incremented by step.

    """
    return step * np.arange(start / step, (stop + step) / step)


def check_uniform_step_width(lst):
    """
    Check to see if a non-uniform step width is used in an list.

    Parameters
    ----------
    lst : list
        List of data points.

    Returns
    -------
    bool
        False if list is non-uniformally spaced.

    """
    start = lst[0]
    stop = lst[-1]
    step = get_minimal_step(lst)

    if step != 0.0 and np.abs((stop - start) / step) > len(lst):
        return False
    return True


def get_minimal_step(lst):
    """
    Return the minimal difference between two consecutive values
    in a list. Used for extracting minimal difference in a
    list with non-uniform spacing.

    Parameters
    ----------
    lst : list
        List of data points.

    Returns
    -------
    step : float
        Non-zero, minimal distance between consecutive data
        points in lst.

    """
    lst1 = np.roll(lst, -1)
    diff = np.abs(np.subtract(lst, lst1))
    step = round(np.min(diff[diff != 0]), 2)

    return step


def _resample_array(y, x0, x1):
    """
    Resample an array (y) which has the same initial spacing
    of another array(x0), based on the spacing of a new
    array(x1).

    Parameters
    ----------
    y : array
        Lineshape array.
    x0 : array
        x array with old spacing.
    x1 : array
        x array with new spacing.

    Returns
    -------
    list
        Interpolated y array.

    """
    # pylint: disable=invalid-name
    interp_fn = interp1d(x0, y, axis=0, fill_value="extrapolate")
    return interp_fn(x1)


def interpolate_arrays(x, array_list):
    """
    Interpolate data points in case a non-uniform step width was used.

    Parameters
    ----------
    x : list
        List of non-uniformally spaced data points.
    array_list : list
        List of arrays to be interpolated according to new x axis.

    Returns
    -------
    x, array_list
        Interpolated x axis and list of arrays

    """
    # pylint: disable=invalid-name
    if not isinstance(array_list, list):
        array_list = [array_list]
    start = x[0]
    stop = x[-1]
    step = get_minimal_step(x)
    if start > stop:
        # pylint: disable=arguments-out-of-order
        new_x = np.flip(safe_arange_with_edges(stop, start, step))
    else:
        new_x = safe_arange_with_edges(start, stop, step)

    output_list = [_resample_array(arr, x, new_x) for arr in array_list]

    return new_x, output_list


def construct_data_key(spectrum):
    """
    Construct a key for the 'data' field of the xps_dict.
    Output example: cycle0_scan0.

    """
    if "loop_no" in spectrum:
        cycle_key = f'cycle{spectrum["loop_no"]}'
    else:
        cycle_key = "cycle0"

    if "scan_no" in spectrum:
        scan_key = f'scan{spectrum["scan_no"]}'
    else:
        scan_key = "scan0"

    return f"{cycle_key}_{scan_key}"


def construct_detector_data_key(spectrum):
    """
    Construct a key for the detector data fields of the xps_dict.
    Output example: 'cycles/Cycle_0/scans/Scan_0'

    """
    if "loop_no" in spectrum:
        cycle_key = f'cycles/Cycle_{spectrum["loop_no"]}'
    else:
        cycle_key = "cycles/Cycle_0"

    if "scan_no" in spectrum:
        scan_key = f'scans/Scan_{spectrum["scan_no"]}'
    else:
        scan_key = "scans/Scan_0"

    key = f"{cycle_key}/{scan_key}"

    if "channel_no" in spectrum:
        key += f'/channels/Channel_{spectrum["channel_no"]}'

    return key


def construct_entry_name(key):
    """Construction entry name."""
    key_parts = key.split("/")
    try:
        # entry example : sample__name_of_scan_region
        entry_name = (
            f'{key_parts[2].split("_", 1)[1]}' f"__" f'{key_parts[4].split("_", 1)[1]}'
        )
    except IndexError:
        entry_name = ""
    return entry_name



"""
Class for reading XPS files from raw VMS data.
"""
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=too-many-lines

import re
from copy import deepcopy
import datetime
from pathlib import Path
from typing import Any, Dict, List, Union

from itertools import groupby
import xarray as xr
import numpy as np

from pynxtools_xps.vms.vamas_data_model import VamasHeader, VamasBlock
from pynxtools_xps.vms.casa_data_model import CasaProcess
# from pynxtools_xps.phi.spe_pro_phi import PhiParser

from pynxtools_xps.reader_utils import (
    XPSMapper,
    construct_entry_name,
    construct_data_key,
    construct_detector_data_key,
    to_snake_case,
    get_minimal_step,
)

UNITS: dict = {
    "instrument/sample_normal_polarangle_tilt": "degree ",
    "instrument/sample_tilt_azimuth": "degree",
    "instrument/sample_rotation_angle": "degree",
    "source/source_analyzer_angle": "degree",
    "source/excitation_energy": "eV",
    "source/particle_charge": "C",
    "analyser/analyzer_take_off_azimuth": "degree",
    "analyser/analyzer_take_off_polar": "degree",
    "analyser/analysis_width_x": "m",
    "analyser/analysis_width_y": "m",
    "analyser/target_bias": "V",
    "analyser/time_correction": "s",
    "analyser/work_function": "eV",
    "energydispersion/pass_energy": "eV",
    "detector/dwell_time": "s",
    "data/start_energy": "eV",
    "data/step_size": "eV",
}


class VamasMapper(XPSMapper):
    """
    Class for restructuring .txt data file from
    Vamas format into python dictionary.
    """

    config_file = "config_vms.json"

    def _select_parser(self):
        """
        Select parser based on the structure of the Vamas file,
        i.e., whether it is regular or irregular.

        Returns
        -------
        VamasParserVMS
            Vamas parser for reading this file structure.

        """
        return VamasParser()

    def construct_data(self):
        """Map VMS format to NXmpes-ready dict."""
        # pylint: disable=duplicate-code
        spectra = deepcopy(self.raw_data)

        self._xps_dict["data"]: Dict[str, Any] = {}

        key_map = {
            "user": [],
            "instrument": [
                "sample_normal_polarangle_tilt",
                "sample_tilt_azimuth",
                "sample_rotation_angle",
            ],
            "source": [
                "source_label",
                "source_analyzer_angle",
            ],
            "beam": ["excitation_energy", "particle_charge"],
            "analyser": [
                "analyzer_take_off_azimuth",
                "analyzer_take_off_polar",
                "analysis_width_x",
                "analysis_width_y",
                "target_bias",
                "time_correction",
                "work_function",
            ],
            "collectioncolumn": [
                "magnification",
            ],
            "energydispersion": [
                "scan_mode",
                "pass_energy",
            ],
            "detector": [
                "signal_mode",
                "dwell_time",
            ],
            "manipulator": [],
            "sample": ["sample_name"],
            "data": [
                "x_label",
                "x_units",
                "y_labels_1",
                "y_units_1",
                "y_labels_2",
                "y_units_2",
                "n_values",
                "start_energy",
                "step_size",
            ],
            "region": [
                "analysis_method",
                "element",
                "group_id",
                "group_name",
                "region",
                "scan_no",
                "scans",
                "spectrum_id",
                "spectrum_type",
                "transition",
                "time_stamp",
            ],
        }

        for spectrum in spectra:
            self._update_xps_dict_with_spectrum(spectrum, key_map)

    def _update_xps_dict_with_spectrum(
        self, spectrum: Dict[str, Any], key_map: Dict[str, str]
    ):
        """Map one spectrum from raw data to NXmpes-ready dict."""
        # pylint: disable=too-many-locals,duplicate-code
        group_parent = f'{self._root_path}/Group_{spectrum["group_name"]}'
        region_parent = f'{group_parent}/regions/Region_{spectrum["spectrum_type"]}'
        instrument_parent = f"{region_parent}/instrument"
        analyser_parent = f"{instrument_parent}/analyser"

        path_map: Dict[str, str] = {
            "user": f"{region_parent}/user",
            "instrument": f"{instrument_parent}",
            "source": f"{instrument_parent}/source",
            "beam": f"{instrument_parent}/beam",
            "analyser": f"{analyser_parent}",
            "collectioncolumn": f"{analyser_parent}/collectioncolumn",
            "energydispersion": f"{analyser_parent}/energydispersion",
            "detector": f"{analyser_parent}/detector",
            "manipulator": f"{instrument_parent}/manipulator",
            "sample": f"{region_parent}/sample",
            "data": f"{region_parent}/data",
            "energy_referencing": f"{region_parent}/calibrations/energy_referencing",
            "peak_fitting": f"{region_parent}/peak_fitting",
            "region": f"{region_parent}",
        }

        used_keys = []

        for grouping, spectrum_keys in key_map.items():
            root = path_map[grouping]
            for spectrum_key in spectrum_keys:
                mpes_key = spectrum_key.rsplit(" ", 1)[0]
                self._xps_dict[f"{root}/{mpes_key}"] = spectrum[spectrum_key]

                unit_key = f"{grouping}/{spectrum_key}"
                units = self._get_units_for_key(unit_key)
                if units:
                    self._xps_dict[f"{root}/{mpes_key}/@units"] = units

        # Write process data
        process_key_map: Dict[str, List[str]] = {
            "energy_referencing": ["alignments"],
            "peak_fitting": ["regions", "components"],
        }

        for grouping, process_key_list in process_key_map.items():
            root = path_map[grouping]
            for spectrum_key in process_key_list:
                try:
                    processes = spectrum[spectrum_key]
                    for i, process in enumerate(processes):
                        process_key = (
                            f"{root}/{spectrum_key}/{spectrum_key.rstrip('s')}{i}"
                        )
                        for key, value in process.dict().items():
                            key = key.replace("_units", "/@units")
                            self._xps_dict[f"{process_key}/{key}"] = value
                    used_keys += [spectrum_key]
                except KeyError:
                    pass

        # Create keys for writing to data and detector
        entry = construct_entry_name(region_parent)
        scan_key = construct_data_key(spectrum)
        detector_data_key_child = construct_detector_data_key(spectrum)
        detector_data_key = f'{path_map["detector"]}/{detector_data_key_child}/counts'

        energy = np.array(spectrum["data"]["x"])
        intensity_raw = np.array(spectrum["data"]["y"])
        intensity_cps = np.array(spectrum["data"]["y_cps"])

        if entry not in self._xps_dict["data"]:
            self._xps_dict["data"][entry] = xr.Dataset()

        # Write averaged cycle data to 'data'.
        all_scan_data = [
            np.array(value)
            for key, value in self._xps_dict["data"][entry].items()
            if scan_key.split("_")[0] in key
        ]

        # Write averaged cycle data to 'data'.
        averaged_scans = np.mean(all_scan_data, axis=0)
        if averaged_scans.size == 1:
            # on first scan in cycle
            averaged_scans = intensity_cps

        try:
            self._xps_dict["data"][entry][scan_key.split("_")[0]] = xr.DataArray(
                data=averaged_scans,
                coords={"energy": energy},
            )
        except ValueError:
            pass

        self._xps_dict["data"][entry][scan_key] = xr.DataArray(
            data=intensity_cps, coords={"energy": energy}
        )
        used_keys += ["data"]

        # Write raw intensities to 'detector'.
        self._xps_dict[detector_data_key] = intensity_raw

        # Write additional keys to region parent.
        for spectrum_key, value in spectrum.items():
            if spectrum_key not in used_keys:
                self._xps_dict[f"{region_parent}/{spectrum_key}"] = value

    def _get_units_for_key(self, unit_key: str):
        """
        Get correct units for a given key.

        Parameters
        ----------
        unit_key : str
           Key of type <mapping>:<spectrum_key>, e.g.
           detector/detector_voltage

        Returns
        -------
        str
            Unit for that unit_key.

        """
        try:
            return re.search(r"\[([A-Za-z0-9_]+)\]", unit_key).group(1)
        except AttributeError:
            try:
                return UNITS[unit_key]
            except KeyError:
                return ""


class VamasParser:
    """A parser for reading vamas files."""

    def __init__(self):
        """Construct the vamas parser.

        Class attributes are a VamasHeader, which stores the vamas header
        attributes, blocks, which store the individual Block objects. Each
        block represents one spectrum, then there are several kinds of
        vamas attribute keys, which are used, depending on how the
        vamas file is formatted.
        """
        self.data: List[str] = []

        self.header = VamasHeader()
        self.blocks: List[VamasBlock] = []

        self.attrs = {
            "common_header": [
                "format_id",
                "institute_id",
                "instrumentModel_id",
                "operator_id",
                "experiment_id",
                "no_comment_lines",
            ],
            "exp_var": ["exp_var_label", "exp_var_unit"],
            "norm_header": [
                "scan_mode",
                "nr_regions",
                "nr_exp_var",
                "unknown_3",
                "unknown_4",
                "unknown_5",
                "unknown_6",
                "no_blocks",
            ],
            "map_header": [
                "scan_mode",
                "nr_regions",
                "nr_positions",
                "nr_x_coords",
                "nr_y_coords",
                "nr_exp_var",
                "unknown_3",
                "unknown_4",
                "unknown_5",
                "unknown_6",
                "no_blocks",
            ],
            "norm_block": [
                "block_id",
                "sample_id",
                "year",
                "month",
                "day",
                "hour",
                "minute",
                "second",
                "no_hrs_in_advance_of_gmt",
                "no_comment_lines",
                "comment_lines",
                "technique",
                "exp_var_value",
                "source_label",
                "source_energy",
                "unknown_1",
                "unknown_2",
                "unknown_3",
                "source_analyzer_angle",
                "unknown_4",
                "analyzer_mode",
                "resolution",
                "magnification",
                "work_function",
                "target_bias",
                "analyzer_width_x",
                "analyzer_width_y",
                "analyzer_take_off_polar_angle",
                "analyzer_azimuth",
                "species_label",
                "transition_label",
                "particle_charge",
                "abscissa_label",
                "abscissa_units",
                "abscissa_start",
                "abscissa_step",
                "no_variables",
                "variable_label1",
                "variable_units1",
                "variable_label2",
                "variable_units2",
                "signal_mode",
                "dwell_time",
                "no_scans",
                "time_correction",
                "sample_angle_tilt",
                "sample_tilt_azimuth",
                "sample_rotation",
                "no_additional_params",
                "param_label_1",
                "param_unit_1",
                "param_value_1",
                "param_label_2",
                "param_unit_2",
                "param_value_2",
                "num_ord_values",
                "min_ord_value_1",
                "max_ord_value_1",
                "min_ord_value_2",
                "max_ord_value_2",
                "data_string",
            ],
            "map_block": [
                "block_id",
                "sample_id",
                "year",
                "month",
                "day",
                "hour",
                "minute",
                "second",
                "no_hrs_in_advance_of_gmt",
                "no_comment_lines",
                "comment_lines",
                "technique",
                "x_coord",
                "y_coord",
                "exp_var_value",
                "source_label",
                "source_energy",
                "unknown_1",
                "unknown_2",
                "unknown_3",
                "fov_x",
                "fov_y",
                "source_analyzer_angle",
                "unknown_4",
                "analyzer_mode",
                "resolution",
                "magnification",
                "work_function",
                "target_bias",
                "analyzer_width_x",
                "analyzer_width_y",
                "analyzer_take_off_polar_angle",
                "analyzer_azimuth",
                "species_label",
                "transition_label",
                "particle_charge",
                "abscissa_label",
                "abscissa_units",
                "abscissa_start",
                "abscissa_step",
                "no_variables",
                "variable_label_1",
                "variable_units_1",
                "variable_label_2",
                "variable_units_2",
                "signal_mode",
                "dwell_time",
                "no_scans",
                "time_correction",
                "sample_angle_tilt",
                "sample_tilt_azimuth",
                "sample_rotation",
                "no_additional_params",
                "param_label_1",
                "param_unit_1",
                "param_value_1",
                "param_label_2",
                "param_unit_2",
                "param_value_2",
                "num_ord_values",
                "min_ord_value_1",
                "max_ord_value_1",
                "min_ord_value_2",
                "max_ord_value_2",
                "data_string",
            ],
        }

    def parse_file(self, file: Union[str, Path]):
        """Parse the vamas file into a list of dictionaries.

        Parameters
        ----------
        file: str
           The location and name of the vamas file to be parsed.
        """
        self._read_lines(file)
        self._parse_header()
        self._parse_blocks()
        return self.build_list()

    def _read_lines(self, file: Union[str, Path]):
        """Read in vamas text file."""
        with open(file, "rb") as vms_file:
            for line in vms_file:
                if line.endswith(b"\r\n"):
                    self.data += [line.decode("utf-8", errors="ignore").strip()]

    def _parse_header(self):
        """Parse the vama header into a VamasHeader object.

        The common_header_attr are the header attributes that are common
        to both types of Vamas format (NORM and MAP).
        Returns
        -------
        None.
        """
        for attr in self.attrs["common_header"]:
            setattr(self.header, attr, self.data.pop(0).strip())
        no_comment_lines = int(self.header.no_comment_lines)
        comments = []
        for _ in range(no_comment_lines):
            comments += [self.data.pop(0)]
        self.header.comment_lines = comments
        self.header.exp_mode = self.data.pop(0).strip()
        if self.header.exp_mode == "NORM":
            for attr in self.attrs["norm_header"]:
                setattr(self.header, attr, self.data.pop(0).strip())
                if attr == "nr_exp_var":
                    self._add_exp_var()

        elif self.header.exp_mode == "MAP":
            for attr in self.attrs["map_header"]:
                setattr(self.header, attr, self.data.pop(0).strip())
                if attr == "nr_exp_var":
                    self._add_exp_var()

        self.header.validate_types()

    def _add_exp_var(self):
        """Add experimental variable to header."""
        for _ in range(int(self.header.nr_exp_var)):
            for attr in self.attrs["exp_var"]:
                setattr(self.header, attr, self.data.pop(0).strip())

    def _parse_blocks(self):
        """Parse all (metadata) of Vamas blocks."""
        for _ in range(int(self.header.no_blocks)):
            self._parse_one_block()

    def _parse_one_block(self):
        """Parse one Vamas Block."""
        if self.header.exp_mode == "NORM":
            self.blocks += [self._parse_norm_block()]
        elif self.header.exp_mode == "MAP":
            self.blocks += [self._parse_map_block()]

    def _parse_norm_block(self):
        """
        Use this method when the NORM keyword is present.

        Returns
        -------
        block : vamas.Block object.
            A block represents one spectrum with its metadata.

        """
        # pylint: disable=too-many-statements
        block = VamasBlock()
        block.block_id = self.data.pop(0).strip()
        block.sample_id = self.data.pop(0).strip()
        block.year = int(self.data.pop(0).strip())
        block.month = int(self.data.pop(0).strip())
        block.day = int(self.data.pop(0).strip())
        block.hour = int(self.data.pop(0).strip())
        block.minute = int(self.data.pop(0).strip())
        block.second = int(self.data.pop(0).strip().split(".")[0])
        block.no_hrs_in_advance_of_gmt = int(self.data.pop(0).strip())
        block.no_comment_lines = int(self.data.pop(0).strip())
        for _ in range(block.no_comment_lines):
            block.comment_lines += [self.data.pop(0)]
        block.technique = self.data.pop(0).strip()
        for _ in range(int(self.header.nr_exp_var)):
            block.exp_var_value = self.data.pop(0).strip()
        block.source_label = self.data.pop(0).strip()
        block.source_energy = float(self.data.pop(0).strip())
        block.unknown_1 = self.data.pop(0).strip()
        block.unknown_2 = self.data.pop(0).strip()
        block.unknown_3 = self.data.pop(0).strip()
        block.source_analyzer_angle = float(self.data.pop(0).strip())
        block.unknown_4 = self.data.pop(0).strip()
        block.analyzer_mode = self.data.pop(0).strip()
        block.resolution = float(self.data.pop(0).strip())
        block.magnification = self.data.pop(0).strip()
        block.work_function = float(self.data.pop(0).strip())
        block.target_bias = float(self.data.pop(0).strip())
        block.analyzer_width_x = float(self.data.pop(0).strip())
        block.analyzer_width_y = float(self.data.pop(0).strip())
        block.analyzer_take_off_polar_angle = float(self.data.pop(0).strip())
        block.analyzer_azimuth = float(self.data.pop(0).strip())
        block.species_label = self.data.pop(0).strip()
        block.transition_label = self.data.pop(0).strip()
        block.particle_charge = int(self.data.pop(0).strip())

        if self.header.scan_mode == "REGULAR":
            block.abscissa_label = self.data.pop(0).strip()
            block.abscissa_units = self.data.pop(0).strip()
            block.abscissa_start = float(self.data.pop(0).strip())
            block.abscissa_step = float(self.data.pop(0).strip())

            block.no_variables = int(self.data.pop(0).strip())
            for var in range(block.no_variables):
                name = "variable_label_" + str(var + 1)
                setattr(block, name, self.data.pop(0).strip())
                name = "variable_units_" + str(var + 1)
                setattr(block, name, self.data.pop(0).strip())

        else:
            block.no_variables = int(self.data.pop(0).strip()) - 1
            block.abscissa_label = self.data.pop(0).strip()
            block.abscissa_units = self.data.pop(0).strip()
            for var in range(block.no_variables):
                name = "variable_label_" + str(var + 1)
                setattr(block, name, self.data.pop(0).strip())
                name = "variable_units_" + str(var + 1)
                setattr(block, name, self.data.pop(0).strip())

        block.signal_mode = self.data.pop(0).strip()
        block.dwell_time = float(self.data.pop(0).strip())
        block.no_scans = int(self.data.pop(0).strip())
        block.time_correction = self.data.pop(0).strip()
        block.sample_angle_tilt = float(self.data.pop(0).strip())
        block.sample_tilt_azimuth = float(self.data.pop(0).strip())
        block.sample_rotation = float(self.data.pop(0).strip())
        block.no_additional_params = int(self.data.pop(0).strip())
        for param in range(block.no_additional_params):
            name = "param_label_" + str(param + 1)
            setattr(block, name, self.data.pop(0))
            name = "param_unit_" + str(param + 1)
            setattr(block, name, self.data.pop(0))
            name = "param_value_" + str(param + 1)
            setattr(block, name, self.data.pop(0))
        block.num_ord_values = int(self.data.pop(0).strip())
        if self.header.scan_mode == "IRREGULAR":
            del self.data[:2]
        for var in range(block.no_variables):
            name = "min_ord_value_" + str(var + 1)
            setattr(block, name, float(self.data.pop(0).strip()))
            name = "max_ord_value_" + str(var + 1)
            setattr(block, name, float(self.data.pop(0).strip()))

        self._add_data_values(block)

        block.validate_types()
        return block

    def _parse_map_block(self):
        """
        Use this method when the MAP keyword is present.

        Returns
        -------
        block : vamas.Block object.
            A block represents one spectrum with its metadata.

        """
        # pylint: disable=too-many-statements
        block = VamasBlock()
        block.block_id = self.data.pop(0).strip()
        block.sample_id = self.data.pop(0).strip()
        block.year = int(self.data.pop(0).strip())
        block.month = int(self.data.pop(0).strip())
        block.day = int(self.data.pop(0).strip())
        block.hour = int(self.data.pop(0).strip())
        block.minute = int(self.data.pop(0).strip())
        block.second = int(self.data.pop(0).strip())
        block.no_hrs_in_advance_of_gmt = int(self.data.pop(0).strip())
        block.no_comment_lines = int(self.data.pop(0).strip())
        for _ in range(block.no_comment_lines):
            self.data.pop(0)
            block.comment_lines += [self.data.pop(0)]
        block.technique = self.data.pop(0).strip()
        block.x_coord = self.data.pop(0).strip()
        block.y_coord = self.data.pop(0).strip()
        block.exp_var_value = self.data.pop(0).strip()
        block.source_label = self.data.pop(0).strip()
        block.source_energy = float(self.data.pop(0).strip())
        block.unknown_1 = self.data.pop(0).strip()
        block.unknown_2 = self.data.pop(0).strip()
        block.unknown_3 = self.data.pop(0).strip()
        block.fov_x = self.data.pop(0).strip()
        block.fov_y = self.data.pop(0).strip()
        block.source_analyzer_angle = float(self.data.pop(0).strip())
        block.unknown_4 = self.data.pop(0).strip()
        block.analyzer_mode = self.data.pop(0).strip()
        block.resolution = float(self.data.pop(0).strip())
        block.magnification = self.data.pop(0).strip()
        block.work_function = float(self.data.pop(0).strip())
        block.target_bias = float(self.data.pop(0).strip())
        block.analyzer_width_x = float(self.data.pop(0).strip())
        block.analyzer_width_y = float(self.data.pop(0).strip())
        block.analyzer_take_off_polar_angle = float(self.data.pop(0).strip())
        block.analyzer_azimuth = float(self.data.pop(0).strip())
        block.species_label = self.data.pop(0).strip()
        block.transition_label = self.data.pop(0).strip()
        block.particle_charge = int(self.data.pop(0).strip())

        if self.header.scan_mode == "REGULAR":
            block.abscissa_label = self.data.pop(0).strip()
            block.abscissa_units = self.data.pop(0).strip()
            block.abscissa_start = float(self.data.pop(0).strip())
            block.abscissa_step = float(self.data.pop(0).strip())

            block.no_variables = int(self.data.pop(0).strip())
            for var in range(block.no_variables):
                name = "variable_label_" + str(var + 1)
                setattr(block, name, self.data.pop(0).strip())
                name = "variable_units_" + str(var + 1)
                setattr(block, name, self.data.pop(0).strip())

        else:
            block.no_variables = int(self.data.pop(0).strip()) - 1
            block.abscissa_label = self.data.pop(0).strip()
            block.abscissa_units = self.data.pop(0).strip()
            for var in range(block.no_variables):
                name = "variable_label_" + str(var + 1)
                setattr(block, name, self.data.pop(0).strip())
                name = "variable_units_" + str(var + 1)
                setattr(block, name, self.data.pop(0).strip())

        block.signal_mode = self.data.pop(0).strip()
        block.dwell_time = float(self.data.pop(0).strip())
        block.no_scans = int(self.data.pop(0).strip())
        block.time_correction = self.data.pop(0).strip()
        block.sample_angle_tilt = float(self.data.pop(0).strip())
        block.sample_tilt_azimuth = float(self.data.pop(0).strip())
        block.sample_rotation = float(self.data.pop(0).strip())
        block.no_additional_params = int(self.data.pop(0).strip())
        for param in range(block.no_additional_params):
            name = "param_label_" + str(param + 1)
            setattr(block, name, self.data.pop(0))
            name = "param_unit_" + str(param + 1)
            setattr(block, name, self.data.pop(0))
            name = "param_value_" + str(param + 1)
            setattr(block, name, self.data.pop(0))
        block.num_ord_values = int(self.data.pop(0).strip())
        if self.header.scan_mode == "IRREGULAR":
            del self.data[:2]
        for var in range(block.no_variables):
            name = "min_ord_value_" + str(var + 1)
            setattr(block, name, float(self.data.pop(0).strip()))
            name = "max_ord_value_" + str(var + 1)
            setattr(block, name, float(self.data.pop(0).strip()))

        self._add_data_values(block)

        return block

    def _add_data_values(self, block: VamasBlock):
        """Add data values to a Vamas data block."""
        if self.header.scan_mode == "REGULAR":
            self._add_regular_data(block)
        elif self.header.scan_mode == "IRREGULAR":
            self._add_irregular_data(block)

    def _add_regular_data(self, block: VamasBlock):
        """Parse data with regularly spaced energy axis."""
        data_dict: Dict[str, np.ndarray] = {}

        start = float(block.abscissa_start)
        step = float(block.abscissa_step)
        num = int(block.num_ord_values / block.no_variables)
        energy = np.array([round(start + i * step, 2) for i in range(num)])

        if block.abscissa_label == "binding energy":
            energy = np.flip(energy)

        setattr(block, "x", energy)

        for var in range(block.no_variables):
            if var == 0:
                name = "y"
            else:
                name = "y" + str(var)

        data_array = np.array(self.data[: block.num_ord_values], dtype=float)

        self.data = self.data[block.num_ord_values :]

        for var in range(block.no_variables):
            max_var = block.no_variables
            if var == 0:
                name = "y"
            else:
                name = "y" + str(var)
            data_array_slice = data_array[var::max_var]
            data_dict[name] = data_array_slice
            setattr(block, name, data_dict[name])

    def _add_irregular_data(self, block: VamasBlock):
        """Parse data with regularly spaced energy axis."""
        data_dict: Dict[str, np.ndarray] = {}

        block_data = np.array(self.data[: block.num_ord_values], dtype=float)

        energy = block_data[:: block.no_variables + 1]
        if block.abscissa_label == "binding energy":
            energy = np.flip(energy)

        setattr(block, "x", energy)
        block.abscissa_start = float(min(energy))
        block.abscissa_step = float(get_minimal_step(energy))

        for var in range(block.no_variables):
            if var == 0:
                name = "y"
            else:
                name = "y" + str(var)

        for var in range(block.no_variables):
            if var == 0:
                name = "y"
            else:
                name = "y" + str(var)
            data_array_slice = block_data[var + 1 :: block.no_variables + 1]
            data_dict[name] = data_array_slice
            setattr(block, name, data_dict[name])

        self.data = self.data[block.num_ord_values :]

    def _get_scan_numbers_for_spectra(self, spectra: List[Dict]):
        """
        For a flat list of spectra dictionaries, group the spectra
        by group name and spectrum type and iteratively give them
        scan numbers.

        Parameters
        ----------
        spectra : list
            List of dicts with each dict containing data and metadata
            for one spectrum.

        Returns
        -------
        flattened_spectra : list
            Same list of dicts, but each spectrum gets a scan number.

        """
        grouped_spectra = [
            list(y)
            for x, y in groupby(
                sorted(spectra, key=lambda x: (x["group_name"], x["spectrum_type"])),
                lambda x: (x["group_name"], x["spectrum_type"]),
            )
        ]

        for group in grouped_spectra:
            for i, spectrum in enumerate(group):
                spectrum["scan_no"] = i

        flattened_spectra = [
            spectrum for group in grouped_spectra for spectrum in group
        ]

        return flattened_spectra

    def handle_header_comments(self, comment_list: List[str]):
        """Handle comments (incl. Casa info) for the header."""
        comments = {}

        special_keys = {
            "Casa Info Follows": self._handle_casa_header,
            "SpecsLab Prodigy": self._handle_prodigy_header,
            # "SOFH": self._handle_phi_header,
        }

        for keyword, handle_func in special_keys.items():
            # if any(keyword in line for line in comment_list):
            indices = [i for i, line in enumerate(comment_list) if keyword in line]

            if indices:
                index = indices[0]
                if keyword == "Casa Info Follows":
                    special_comments = comment_list[index]
                    comment_list = comment_list[index + 1 :]

                if keyword == "SpecsLab Prodigy":
                    special_comments = comment_list[index]
                    comment_list = comment_list[index + 1 :]

                if keyword == "SOFH":
                    end_index = [
                        i for i, line in enumerate(comment_list) if "EOFH" in line
                    ][0]
                    special_comments = comment_list[index : end_index + 1]  # type: ignore[assignment]
                    del comment_list[index : end_index + 1]

                comments.update(handle_func(special_comments))

        # Handle non-special comments.
        for line in comment_list:
            for sep in ("=", ":"):
                try:
                    key, value = [part.strip(" ") for part in line.split(sep, 1)]
                    comments[to_snake_case(key)] = value
                except ValueError:
                    continue

        return comments

    def _handle_casa_header(self, comment_line: str):
        """Get information about CasaXPS version."""
        return {
            "casa_version": comment_line.split("Casa Info Follows CasaXPS Version")[
                1
            ].strip()
        }

    def _handle_prodigy_header(self, comment_line: str):
        """Get information about SpecsLab Prodigy version."""
        return {"prodigy_version": comment_line.split("Version")[1].strip()}

    # =============================================================================
    #     def _handle_phi_header(self, comment_list: List[str]):
    #         """Get metadta from Phi system."""
    #         phi_parser = PhiParser()
    #         phi_parser.parse_header_into_metadata(comment_list)
    #
    #         phi_comments = phi_parser.metadata.dict()
    #
    #         regions = phi_parser.parse_spectral_regions(comment_list)
    #         areas = phi_parser.parse_spatial_areas(comment_list)
    #
    #         for region in regions:
    #             for area in areas:
    #                 concatenated = {**region.dict(), **area.dict()}
    #
    #             phi_comments.update(concatenated)
    #
    #         return phi_comments
    # =============================================================================

    def handle_block_comments(self, comment_list: List[str]):
        """Handle comments (incl. Casa fitting) for each block."""
        comments = {}

        if "Casa Info Follows" in comment_list[0]:
            # Get all processing and fitting data from Casa comments.
            casa = CasaProcess()
            casa_data = casa.process_comments(comment_list)

            comments.update(casa_data)

            no_of_casa_lines = 1

            for number in (
                "n_alignments",
                "n_unknown_processes",
                "n_regions",
                "n_components",
            ):
                occurence = getattr(casa, number)
                no_of_casa_lines += 1
                if occurence >= 1:
                    no_of_casa_lines += occurence

            non_casa_comments = comment_list[no_of_casa_lines:]

        else:
            non_casa_comments = comment_list

        for line in non_casa_comments:
            for sep in ("=", ":"):
                try:
                    key, value = [part.strip(" ") for part in line.split("=", 1)]
                    comments[to_snake_case(key)] = value
                except ValueError:
                    continue
        return comments

    def build_list(self):
        """
        Construct a list of dictionaries from the Vamas objects

        Returns
        -------
        List
            Each list element is a dictionary with the data and
            metadata of one spectrum.

        """
        group_id = -1
        temp_group_name = ""
        spectra = []

        header_dict = {to_snake_case(k): v for (k, v) in self.header.dict().items()}
        del header_dict["comment_lines"]
        header_dict.update(self.handle_header_comments(self.header.comment_lines))

        for spectrum_id, block in enumerate(self.blocks):
            group_name = block.sample_id
            # This set of conditions detects if the group name has changed.
            # If it has, then it increments the group_idx.
            if group_name != temp_group_name:
                temp_group_name = group_name
                group_id += 1

            spectrum_type = str(block.species_label + block.transition_label)

            settings = {
                "region": block.block_id,
                "sample_name": block.sample_id,
                "analysis_method": block.technique,
                "source_label": block.source_label,
                "excitation_energy": block.source_energy,
                "source_analyzer_angle": block.source_analyzer_angle,
                "scan_mode": block.analyzer_mode,
                "pass_energy": block.resolution,
                "magnification": block.magnification,
                "work_function": block.work_function,
                "target_bias": block.target_bias,
                "analysis_width_x": block.analyzer_width_x,
                "analysis_width_y": block.analyzer_width_y,
                "analyzer_take_off_polar": block.analyzer_take_off_polar_angle,
                "analyzer_take_off_azimuth": block.analyzer_azimuth,
                "element": block.species_label,
                "transition": block.transition_label,
                "particle_charge": block.particle_charge,
                "x_label": block.abscissa_label,
                "x_units": block.abscissa_units,
                "start_energy": block.abscissa_start,
                "step_size": block.abscissa_step,
                "y_labels_1": block.variable_label_1,
                "y_units_1": block.variable_units_1,
                "y_labels_2": block.variable_label_2,
                "y_units_2": block.variable_units_2,
                "signal_mode": block.signal_mode,
                "dwell_time": block.dwell_time,
                "time_correction": block.time_correction,
                "sample_normal_polarangle_tilt": block.sample_angle_tilt,
                "sample_tilt_azimuth": block.sample_tilt_azimuth,
                "sample_rotation_angle": block.sample_rotation,
                "n_values": int(block.num_ord_values / block.no_variables),
            }
            settings.update(header_dict)

            comment_dict = self.handle_block_comments(block.comment_lines)
            settings.update(comment_dict)

            # Convert the native time format to the datetime string
            # in the ISO 8601 format
            tzinfo = datetime.timezone(
                datetime.timedelta(hours=block.no_hrs_in_advance_of_gmt)
            )
            try:
                date_time = datetime.datetime(
                    block.year,
                    block.month,
                    block.day,
                    block.hour,
                    block.minute,
                    block.second,
                    tzinfo=tzinfo,
                )
            except ValueError:
                date_time = datetime.datetime.min

            data = {"x": block.x}
            for var in range(int(block.no_variables)):
                if var == 0:
                    key = "y"

                    data["y"] = getattr(block, "y")

                    if block.variable_label_1 in ["Intensity", "counts"]:
                        y_cps = [np.round(y / block.dwell_time, 2) for y in block.y]
                        data["y_cps"] = np.array(y_cps)

                else:
                    key = "y" + str(var)
                    data[key] = getattr(block, key)

            spec_dict = {
                "time_stamp": date_time,
                "group_name": group_name,
                "group_id": group_id,
                "spectrum_type": spectrum_type,
                "spectrum_id": spectrum_id,
                "scans": block.no_scans,
                "data": data,
            }
            spec_dict.update(settings)
            spectra += [spec_dict]

        spectra = self._get_scan_numbers_for_spectra(spectra)

        return spectra



"""
Class for reading XPS files from TXT export of Scienta.
"""
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=too-many-lines

import re
import copy
from datetime import datetime
import pytz
import xarray as xr
import numpy as np

from pynxtools_xps.reader_utils import (
    XPSMapper,
    construct_entry_name,
    construct_data_key,
    construct_detector_data_key,
)


class TxtMapperScienta(XPSMapper):
    """
    Class for restructuring .txt data file from
    Scienta TXT export into python dictionary.
    """

    config_file = "config_txt_scienta.json"

    def _select_parser(self):
        """
        Select Scienta TXT parser.
        Currently, there is only one parser.

        Returns
        -------
        ScientaTxtHelper
            Parser for reading .txt file exported by Scienta.

        """
        return ScientaTxtHelper()

    def construct_data(self):
        """Map TXT format to NXmpes-ready dict."""
        # pylint: disable=duplicate-code
        spectra = copy.deepcopy(self.raw_data)

        self._xps_dict["data"]: dict = {}

        key_map = {
            "file_info": ["data_file", "sequence_file"],
            "user": [
                "user_initials",
            ],
            "instrument": [
                "instrument_name",
                "vendor",
            ],
            "source": [],
            "beam": ["excitation_energy"],
            "analyser": [],
            "collectioncolumn": [
                "lens_mode",
            ],
            "energydispersion": [
                "acquisition_mode",
                "pass_energy",
            ],
            "detector": [
                "detector_first_x_channel",
                "detector_first_y_channel",
                "detector_last_x_channel",
                "detector_last_y_channel",
                "detector_mode",
                "dwell_time",
            ],
            "manipulator": [],
            "calibration": [],
            "sample": ["sample_name"],
            "data": [
                "x_units",
                "energy_axis",
                "energy_type",
                "step_size",
            ],
            "region": [
                "center_energy",
                "energy_scale",
                "energy_size",
                "no_of_scans",
                "region_id",
                "spectrum_comment",
                "start_energy",
                "stop_energy",
                "time_stamp",
            ],
            # 'unused': [
            #     'energy_unit',
            #     'number_of_slices',
            #     'software_version',
            #     'spectrum_comment',
            #     'start_date',
            #     'start_time',
            #     'time_per_spectrum_channel'
            # ]
        }

        for spectrum in spectra:
            self._update_xps_dict_with_spectrum(spectrum, key_map)

    def _update_xps_dict_with_spectrum(self, spectrum, key_map):
        """
        Map one spectrum from raw data to NXmpes-ready dict.

        """
        # pylint: disable=too-many-locals,duplicate-code
        group_parent = f'{self._root_path}/Region_{spectrum["spectrum_type"]}'
        region_parent = f'{group_parent}/regions/Region_{spectrum["region_name"]}'
        file_parent = f"{region_parent}/file_info"
        instrument_parent = f"{region_parent}/instrument"
        analyser_parent = f"{instrument_parent}/analyser"

        path_map = {
            "file_info": f"{file_parent}",
            "user": f"{region_parent}/user",
            "instrument": f"{instrument_parent}",
            "source": f"{instrument_parent}/source",
            "beam": f"{instrument_parent}/beam",
            "analyser": f"{analyser_parent}",
            "collectioncolumn": f"{analyser_parent}/collectioncolumn",
            "energydispersion": f"{analyser_parent}/energydispersion",
            "detector": f"{analyser_parent}/detector",
            "manipulator": f"{instrument_parent}/manipulator",
            "calibration": f"{instrument_parent}/calibration",
            "sample": f"{region_parent}/sample",
            "data": f"{region_parent}/data",
            "region": f"{region_parent}",
        }

        for grouping, spectrum_keys in key_map.items():
            root = path_map[str(grouping)]
            for spectrum_key in spectrum_keys:
                try:
                    units = re.search(r"\[([A-Za-z0-9_]+)\]", spectrum_key).group(1)
                    mpes_key = spectrum_key.rsplit(" ", 1)[0]
                    self._xps_dict[f"{root}/{mpes_key}/@units"] = units
                    self._xps_dict[f"{root}/{mpes_key}"] = spectrum[spectrum_key]
                except AttributeError:
                    mpes_key = spectrum_key
                    self._xps_dict[f"{root}/{mpes_key}"] = spectrum[spectrum_key]

        # Create keys for writing to data and detector
        entry = construct_entry_name(region_parent)
        scan_key = construct_data_key(spectrum)
        detector_data_key_child = construct_detector_data_key(spectrum)
        detector_data_key = f'{path_map["detector"]}/{detector_data_key_child}/counts'

        # Write raw data to detector.
        self._xps_dict[detector_data_key] = spectrum["data"]["y"]

        # If multiple spectra exist to entry, only create a new
        # xr.Dataset if the entry occurs for the first time.
        if entry not in self._xps_dict["data"]:
            self._xps_dict["data"][entry] = xr.Dataset()

        energy = np.array(spectrum["data"]["x"])
        intensity = spectrum["data"]["y"]

        # Write to data in order: scan, cycle, channel

        # Write averaged cycle data to 'data'.
        all_scan_data = [
            value
            for key, value in self._xps_dict["data"][entry].items()
            if scan_key.split("_")[0] in key
        ]
        averaged_scans = np.mean(all_scan_data, axis=0)
        if averaged_scans.size == 1:
            # on first scan in cycle
            averaged_scans = intensity

        self._xps_dict["data"][entry][scan_key.split("_")[0]] = xr.DataArray(
            data=averaged_scans,
            coords={"energy": energy},
        )

        # Write scan data to 'data'.
        self._xps_dict["data"][entry][scan_key] = xr.DataArray(
            data=intensity, coords={"energy": energy}
        )

        # Write channel data to 'data'.
        channel_key = f"{scan_key}_chan0"
        self._xps_dict["data"][entry][channel_key] = xr.DataArray(
            data=intensity, coords={"energy": energy}
        )


class ScientaTxtHelper:
    """Parser for Scienta TXT exports."""

    # pylint: disable=too-few-public-methods

    def __init__(self):
        self.lines: list = []
        self.spectra: list = []
        self.no_of_regions = 0

        keys_map = {
            "Number of Regions": "no_of_regions",
            "Version": "software_version",
            "Lens Mode": "lens_mode",
            "Pass Energy": "pass_energy",
            "Number of Sweeps": "no_of_scans",
            "Excitation Energy": "excitation_energy",
            "Energy Scale": "energy_scale",
            "Acquisition Mode": "acquisition_mode",
            "Energy Unit": "x_units",
            "Center Energy": "center_energy",
            "Low Energy": "start_energy",
            "High Energy": "stop_energy",
            "Energy Step": "step_size",
            "Step Time": "dwell_time",
            "Number of Slices": "number_of_slices",
            "File": "data_file",
            "Sequence": "sequence_file",
            "Spectrum Name": "spectrum_type",
            "Instrument": "instrument_name",
            "Location": "vendor",
            "User": "user_initials",
            "Sample": "sample_name",
            "Comments": "spectrum_comment",
            "Date": "start_date",
            "Time": "start_time",
            "Time per Spectrum Channel": "time_per_spectrum_channel",
            "DetectorMode": "detector_mode",
        }

        self.region_keys_map = {
            "Region Name": "region_name",
            "name": "energy_type",
            "size": "energy_size",
            "scale": "energy_axis",
        }

        detector_map = {
            "Detector First X-Channel": "detector_first_x_channel",
            "Detector Last X-Channel": "detector_last_x_channel",
            "Detector First Y-Channel": "detector_first_y_channel",
            "Detector Last Y-Channel": "detector_last_y_channel",
        }

        lens_mode_map = {"Transmission": "fixed analyzer transmission"}

        self.key_maps = [
            keys_map,
            self.region_keys_map,
            detector_map,
            lens_mode_map,
        ]

        self.value_map = {
            "no_of_regions": self._change_no_of_regions,
            "x_units": self._change_energy_type,
            "energy_axis": [self._separate_dimension_scale, np.array],
            "energy_scale": self._change_energy_type,
            "acquisition_mode": self._change_scan_mode,
        }

    def parse_file(self, file):
        """
        Parse the file's data and metadata into a flat
        list of dictionaries.


        Parameters
        ----------
        file : str
            Filepath of the TXT file to be read.

        Returns
        -------
        self.spectra
            Flat list of dictionaries containing one spectrum each.

        """
        self._read_lines(file)
        self._parse_header()

        for region_id in range(1, self.no_of_regions + 1):
            self._parse_region(region_id)

        return self.spectra

    def _read_lines(self, file):
        """
        Read all lines from the input txt files.


        Parameters
        ----------
        file : str
            Filepath of the TXT file to be read.

        Returns
        -------
        None.

        """
        with open(file, encoding="utf-8") as txt_file:
            for line in txt_file:
                self.lines += [line]

    def _parse_header(self):
        """
        Parse header with information about the software version
        and the number of spectra in the file.

        Returns
        -------
        None.

        """
        n_headerlines = 4
        header = self.lines[:n_headerlines]
        self.lines = self.lines[n_headerlines:]

        for line in header:
            key, value = self._get_key_value_pair(line)
            if key:
                value = self._re_map_values(key, value)
                setattr(self, key, value)

    def _parse_region(self, region_id):
        """
        Parse data from one region (i.e., one measured spectrum)
        into a dictioanry and append to all spectra.

        Parameters
        ----------
        region_id : int
            Number of the region in the file.

        Returns
        -------
        None.

        """

        region_data = {
            "region_id": region_id,
            "data": {"x": [], "y": []},
        }

        begin_region = False
        begin_info = False
        begin_data = False

        for line in self.lines:
            if line.startswith(f"[Region {region_id}"):
                begin_region = True
            if line.startswith(f"[Info {region_id}"):
                begin_info = True
            if line.startswith(f"[Data {region_id}"):
                begin_data = True
                continue

            if line.startswith("\n"):
                begin_region = False
                begin_info = False
                begin_data = False

            if begin_region:
                # Read region meta data.
                key, value = self._get_key_value_pair(line)
                if "Dimension" in key:
                    key = self._re_map_keys(key)
                    key = self.region_keys_map[key.rsplit(" ")[-1]]
                    key = key.rsplit(" ")[-1]
                    value = self._re_map_values(key, value)
                    if self._check_valid_value(value):
                        region_data[key] = value

            if begin_info:
                # Read instrument meta data for this region.
                key, value = self._get_key_value_pair(line)
                if self._check_valid_value(value):
                    region_data[key] = value

            if begin_data:
                # Read XY data for this region.
                [energy, intensity] = [float(s) for s in line.split(" ") if s != ""]

                region_data["data"]["x"].append(energy)
                region_data["data"]["y"].append(intensity)

        region_data["data"]["x"] = np.array(region_data["data"]["x"])
        region_data["data"]["y"] = np.array(region_data["data"]["y"])

        # Convert date and time to ISO8601 date time.
        start_date = region_data["start_date"]
        start_time = region_data["start_time"]
        region_data["time_stamp"] = self._construct_date_time(start_date, start_time)

        self.spectra.append(region_data)

    def _check_valid_value(self, value):
        """
        Check if a string or an array is empty.

        Parameters
        ----------
        value : obj
            For testing, this can be a str or a np.ndarray.

        Returns
        -------
        bool
            True if the string or np.ndarray is not empty.

        """
        if isinstance(value, str) and value:
            return True
        if isinstance(value, np.ndarray) and value.size != 0:
            return True
        return False

    def _get_key_value_pair(self, line):
        """
        Split the line at the '=' sign and return a
        key-value pair. The values are mapped according
        to the desired format.

        Parameters
        ----------
        line : str
            One line from the input file.

        Returns
        -------
        key : str
            Anything before the '=' sign, mapped to the desired
            key format.
        value : obj
            Anything after the '=' sign, mapped to the desired
            value format and type.

        """
        try:
            [key, value] = line.split("=")
            key = self._re_map_keys(key)
            value = self._re_map_values(key, value)
            return key, value

        except ValueError:
            key, value = "", ""
            return key, value

    def _re_map_keys(self, key):
        """
        Map the keys returned from the file to the preferred keys for
        the parser output.

        """
        maps = {}
        for key_map in self.key_maps:
            maps.update(key_map)

        keys = list(maps.keys())

        if key in keys:
            key = maps[key]

        return key

    def _change_no_of_regions(self, no_of_regions):
        """
        Make no_of_regions an integer.
        Maps e.g. from '0001' string to 1.

        Parameters
        ----------
        no_of_regions : str
            No. of regions in the data file,
            read from the first line.

        Returns
        -------
        int
            Ineger value of the number of regions.

        """
        return int(no_of_regions)

    def _change_energy_type(self, energy):
        """
        Change the strings for energy type to the preferred format.

        """
        if "Binding" in energy:
            return "binding energy"
        if "Kinetic" in energy:
            return "kinetic energy"
        return None

    def _separate_dimension_scale(self, scale):
        return [float(s) for s in scale.split(" ")]

    def _change_scan_mode(self, acquisition_mode):
        """
        Change the strings for acquisition mode type to
        the allowed values in NXmpes.

        """
        if acquisition_mode == "Fixed":
            return "fixed"
        if acquisition_mode == "Swept":
            return "sweep"
        return None

    def _construct_date_time(self, date, time):
        """
        Convert the native time format to the datetime string
        in the ISO 8601 format: '%Y-%b-%dT%H:%M:%S.%fZ'.

        """
        date_time = datetime.combine(
            datetime.strptime(date, "%Y-%m-%d"),
            datetime.strptime(time, "%H:%M:%S").time(),
        )

        localtz = pytz.timezone("Europe/Berlin")
        date_time = localtz.localize(date_time)

        return date_time.isoformat()

    def _re_map_values(self, input_key, value):
        """
        Map the values returned from the file to the preferred format for
        the parser output.

        """
        try:
            value = value.rstrip("\n")
        except AttributeError:
            pass

        keys = list(self.value_map.keys())

        for k in keys:
            if k in input_key:
                if not isinstance(self.value_map[k], list):
                    map_methods = [self.value_map[k]]
                else:
                    map_methods = self.value_map[k]
                for method in map_methods:
                    value = method(value)
        return value



"""
Data model for Vamas ISO standard.
"""
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=too-many-instance-attributes

from dataclasses import dataclass, field
from pynxtools_xps.reader_utils import XpsDataclass


@dataclass
class VamasHeader(XpsDataclass):
    """An object to store the Vamas header information."""

    format_id: str = (
        "VAMAS Surface Chemical Analysis Standard Data Transfer Format 1988 May 4"
    )
    institute_id: str = "Not Specified"
    instrument_model_id: str = "Not Specified"
    operator_id: str = "Not Specified"
    experiment_id: str = "Not Specified"
    no_comment_lines: str = "2"
    comment_lines: list = field(
        default_factory=list
    )  # ["Casa Info Follows CasaXPS Version 2.3.22PR1.0\n0"]
    exp_mode: str = "NORM"
    scan_mode: str = "REGULAR"
    nr_regions: str = "0"
    nr_exp_var: str = "1"
    exp_var_label: str = "Exp Variable"
    exp_var_unit: str = "d"
    unknown_3: str = "0"
    unknown_4: str = "0"
    unknown_5: str = "0"
    unknown_6: str = "0"
    no_blocks: str = "1"


@dataclass
class VamasBlock(XpsDataclass):
    """An object to store a block of spectrum data and meta-data."""

    block_id: str = ""
    sample_id: str = ""
    year: int = 0
    month: int = 0
    day: int = 0
    hour: int = 0
    minute: int = 0
    second: int = 0
    no_hrs_in_advance_of_gmt: int = 0
    no_comment_lines: int = 0
    # This list should contain one element per for each
    # line in the comment block
    comment_lines: list = field(default_factory=list)
    technique: str = ""
    exp_var_value: str = ""
    source_label: str = ""
    source_energy: float = 0.0
    unknown_1: str = "0"
    unknown_2: str = "0"
    unknown_3: str = "0"
    source_analyzer_angle: float = 0.0
    unknown_4: str = "180"
    analyzer_mode: str = ""
    resolution: float = 0.0
    magnification: str = "1"
    work_function: float = 0.0
    target_bias: float = 0.0
    # analyser slit length divided by the magnification
    # of the analyser transfer lens
    analyzer_width_x: float = 0.0
    analyzer_width_y: float = 0.0
    # degrees from upward z-direction,
    # defined by the sample stage
    analyzer_take_off_polar_angle: float = 0.0
    analyzer_azimuth: float = 0.0
    species_label: str = ""
    transition_label: str = ""
    particle_charge: int = -1
    abscissa_label: str = "kinetic energy"
    abscissa_units: str = "eV"
    abscissa_start: float = 0.0
    abscissa_step: float = 0.0
    no_variables: int = 2
    variable_label_1: str = "counts"
    variable_units_1: str = "d"
    variable_label_2: str = "Transmission"
    variable_units_2: str = "d"
    signal_mode: str = "pulse counting"
    dwell_time: float = 0.0
    no_scans: int = 0
    time_correction: str = "0"
    # degrees from upward z-direction,
    # defined by the sample stage
    sample_angle_tilt: float = 0.0
    # degrees clockwise from the y-direction towards the
    # operator, defined by the sample stage
    sample_tilt_azimuth: float = 0.0
    sample_rotation: float = 0.0
    no_additional_params: int = 2
    param_label_1: str = "ESCAPE DEPTH TYPE"
    param_unit_1: str = "d"
    param_value_1: str = "0"
    param_label_2: str = "MFP Exponent"
    param_unit_2: str = "d"
    param_value_2: str = "0"
    num_ord_values: int = 0
    min_ord_value_1: float = 0.0
    max_ord_value_1: float = 0.0
    min_ord_value_2: float = 0.0
    max_ord_value_2: float = 0.0
    data_string: str = ""



"""
Parser for reading XPS (X-ray Photoelectron Spectroscopy) data from
Specs Lab Prodigy XML export, to be passed to mpes nxdl
(NeXus Definition Language) template.
"""
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import xml.etree.ElementTree as EmtT
from typing import Tuple, List, Any
import copy
import xarray as xr
import numpy as np

from pynxtools_xps.reader_utils import (
    XPSMapper,
    construct_entry_name,
)


class XmlMapperSpecs(XPSMapper):
    """
    Class for restructuring xml data file from
    specs vendor into python dictionary.
    """

    config_file = "config_xml_specs.json"

    def _select_parser(self):
        """
        Select Specs XML parser.
        Currently, there is only one parser.
        Returns
        -------
        XmlParserSpecs
            Parser for reading .xml file exported by SpecsLab2.

        """
        return XmlParserSpecs()

    def parse_file(self, file, **kwargs):
        """
        Parse the file using the parser that fits the Prodigy SLE version.
        Returns flat list of dictionaries containing one spectrum each.

        """
        self.file = file
        self.parser = self._select_parser()
        metadata_dict, self.raw_data = self.parser.parse_file(file, **kwargs)

        self._xps_dict = {**self._xps_dict, **metadata_dict}

        file_key = f"{self._root_path}/File"
        self._xps_dict[file_key] = file

        self.construct_data()

        return self.data_dict

    # pylint: disable=too-many-locals,too-many-statements, too-many-branches
    def construct_data(self):
        """
        Construct the Binding Energy and separate the counts for
        different detectors and finally sum up all the counts for
        to find total electron counts.
        """
        spectra = copy.deepcopy(self.raw_data)
        self._xps_dict["data"]: dict = {}

        for entry, spectrum in spectra.items():
            raw_data = spectrum["raw_data"]
            mcd_num = int(raw_data["mcd_num"])

            curves_per_scan = raw_data["curves_per_scan"]
            values_per_curve = raw_data["values_per_curve"]
            values_per_scan = int(curves_per_scan * values_per_curve)
            mcd_head = int(raw_data["mcd_head"])
            mcd_tail = int(raw_data["mcd_tail"])
            excitation_energy = raw_data["excitation_energy"]
            scan_mode = raw_data["scan_mode"]
            kinetic_energy = raw_data["kinetic_energy"]
            scan_delta = raw_data["scan_delta"]
            pass_energy = raw_data["pass_energy"]
            kinetic_energy_base = raw_data["kinetic_energy_base"]
            # Adding one unit to the binding_energy_upper is added as
            # electron comes out if energy is one unit higher
            binding_energy_upper = (
                excitation_energy - kinetic_energy + kinetic_energy_base + 1
            )

            mcd_energy_shifts = raw_data["mcd_shifts"]
            mcd_energy_offsets = []
            offset_ids = []

            # consider offset values for detector with respect to
            # position at +16 which is usually large and positive value
            for mcd_shift in mcd_energy_shifts:
                mcd_energy_offset = (mcd_energy_shifts[-1] - mcd_shift) * pass_energy
                mcd_energy_offsets.append(mcd_energy_offset)
                offset_id = round(mcd_energy_offset / scan_delta)
                offset_ids.append(int(offset_id - 1 if offset_id > 0 else offset_id))

            # Skiping entry without count data
            if not mcd_energy_offsets:
                continue
            mcd_energy_offsets = np.array(mcd_energy_offsets)
            # Putting energy of the last detector as a highest energy
            starting_eng_pnts = binding_energy_upper - mcd_energy_offsets
            ending_eng_pnts = starting_eng_pnts - values_per_scan * scan_delta

            channeltron_eng_axes = np.zeros((mcd_num, values_per_scan))
            for ind in np.arange(len(channeltron_eng_axes)):
                channeltron_eng_axes[ind, :] = np.linspace(
                    starting_eng_pnts[ind], ending_eng_pnts[ind], values_per_scan
                )

            channeltron_eng_axes = np.round_(channeltron_eng_axes, decimals=8)
            # construct ultimate or incorporated energy axis from
            # lower to higher energy
            scans = list(raw_data["scans"].keys())

            # Check whether array is empty or not
            if not scans:
                continue
            if not raw_data["scans"][scans[0]].any():
                continue
            # Sorting in descending order
            binding_energy = channeltron_eng_axes[-1, :]

            self._xps_dict["data"][entry] = xr.Dataset()

            for scan_nm in scans:
                channel_counts = np.zeros((mcd_num + 1, values_per_scan))
                # values for scan_nm corresponds to the data for each
                # "scan" in individual CountsSeq
                scan_counts = raw_data["scans"][scan_nm]

                if scan_mode == "FixedAnalyzerTransmission":
                    for row in np.arange(mcd_num):
                        count_on_row = scan_counts[row::mcd_num]
                        # Reverse counts from lower to higher
                        # BE as in BE_eng_axis
                        count_on_row = count_on_row[mcd_head:-mcd_tail]

                        channel_counts[row + 1, :] = count_on_row
                        channel_counts[0, :] += count_on_row

                        # Storing detector's raw counts
                        self._xps_dict["data"][entry][
                            f"{scan_nm}_chan_{row}"
                        ] = xr.DataArray(
                            data=channel_counts[row + 1, :],
                            coords={"energy": binding_energy},
                        )

                        # Storing callibrated and after accumulated each scan counts
                        if row == mcd_num - 1:
                            self._xps_dict["data"][entry][scan_nm] = xr.DataArray(
                                data=channel_counts[0, :],
                                coords={"energy": binding_energy},
                            )
                else:
                    for row in np.arange(mcd_num):
                        start_id = offset_ids[row]
                        count_on_row = scan_counts[start_id::mcd_num]
                        count_on_row = count_on_row[0:values_per_scan]
                        channel_counts[row + 1, :] = count_on_row

                        # shifting and adding all the curves.
                        channel_counts[0, :] += count_on_row

                        # Storing detector's raw counts
                        self._xps_dict["data"][entry][
                            f"{scan_nm}_chan{row}"
                        ] = xr.DataArray(
                            data=channel_counts[row + 1, :],
                            coords={"energy": binding_energy},
                        )

                        # Storing callibrated and after accumulated each scan counts
                        if row == mcd_num - 1:
                            self._xps_dict["data"][entry][scan_nm] = xr.DataArray(
                                data=channel_counts[0, :],
                                coords={"energy": binding_energy},
                            )


class XmlParserSpecs:
    """Parser for SpecsLab2 XML data"""

    def __init__(self) -> None:
        self.metadata_dict: dict = {}
        self.entry_to_data: dict = {}
        self._root_path = "/ENTRY[entry]"
        self.tail_part_frm_struct = ""
        self.tail_part_frm_othr = ""
        self.child_nm_reslvers = "__child_name_resolver__"

    def parse_file(self, file):
        """Start parsing process and parse children recursively.

        Parameters
        ----------
        """
        root_element = EmtT.parse(file).getroot()
        root_element.attrib[self.child_nm_reslvers] = []
        child_num = len(root_element)
        parent_path = self._root_path
        skip_child = -1

        child_elmt_ind = 0
        while child_num > 0:
            self.pass_child_through_parsers(
                root_element, parent_path, child_elmt_ind, skip_child
            )

            child_num -= 1
            child_elmt_ind += 1

        self.collect_raw_data_to_construct_data()

        return self.metadata_dict, self.entry_to_data

    def pass_child_through_parsers(
        self,
        element_: EmtT.Element,
        parent_path: str,
        child_elmt_ind: int,
        skip_child: int,
    ) -> None:
        """
        Parse the element to parser according to element tag.
        Parameters
        ----------
        element_ : xml element to parse
        parent_path : Xpath of the parent element where the element_ belongs
        child_elmt_ind : Index of child element to track the children.
        skip_child : Tack the children who will be skipped to pass to
                     the parser
        Returns
        -------
        None
        """

        name_val_elmt_tag = ["ulong", "double", "string", "boolean", "enum", "any"]

        parent_element = element_
        element = parent_element[child_elmt_ind]
        element.attrib["__parent__"] = parent_element  # type: ignore[assignment]
        element.attrib["__odr_siblings__"] = child_elmt_ind  # type: ignore[assignment]

        if self.child_nm_reslvers not in parent_element.attrib.keys():
            parent_element.attrib[self.child_nm_reslvers] = []  # type: ignore[assignment]

        elmt_tag = element.tag

        if child_elmt_ind <= skip_child:
            pass

        elif elmt_tag == "sequence":
            self.parse_sequence(element, parent_path)

        elif elmt_tag == "struct":
            self.parse_struct(element, parent_path)

        elif elmt_tag in name_val_elmt_tag:
            self.last_element_parser(element, parent_path)

        else:
            raise TypeError("Needs to parse to different type of parser")

    def parse_sequence(self, element_: EmtT.Element, parent_path: str) -> None:
        """
        Parameters
        ----------
        element_ : Element with sequence tag
        parent_path : Xpath of the parent element where the element_ belongs

        Returns
        -------
        None
        """

        child_num = len(element_)
        elmt_attr = element_.attrib

        section_nm_reslvr = ""
        key_name = "name"
        if key_name in elmt_attr.keys():
            section_nm_reslvr = f"{elmt_attr[key_name]}"
            section_nm_reslvr = self.check_for_siblings_with_same_name(
                section_nm_reslvr, element_
            )

            parent_path = f"{parent_path}/{section_nm_reslvr}"

        child_elmt_ind = 0
        while child_num > 0:
            self.pass_child_through_parsers(
                element_, parent_path, child_elmt_ind, skip_child=-1
            )
            child_num -= 1
            child_elmt_ind += 1

    def struct_fc_name_sc_value(self, element_, first_child, parent_path, skip_child):
        """Struct representing parameter with first child (fc) 'name'
        and second child(sc) 'value'."""
        section_nm_reslvr = ""
        units = ["mV", "deg", "W", "kV", "ns"]

        section_nm_reslvr = self.restructure_value(first_child.text, first_child.tag)
        section_nm_reslvr = self.check_for_siblings_with_same_name(
            section_nm_reslvr, element_
        )
        skip_child += 1
        # Separating the units
        for unit in units:
            if f"_[{unit}]" in section_nm_reslvr:
                section_nm_reslvr, _ = section_nm_reslvr.split("_")
                self.metadata_dict[
                    f"{parent_path}/" f"{section_nm_reslvr}/@unit"
                ] = unit

        parent_path, self.tail_part_frm_struct = self.check_last_part_repetition(
            parent_path, self.tail_part_frm_struct, section_nm_reslvr
        )

        return parent_path, skip_child

    def struct_fc_name_sc_string(self, element_, first_child, parent_path, skip_child):
        """Struct representing parameter with first child(fc) 'name'
        and first child(fc) having 'string'."""

        elmt_attr = element_.attrib
        key_type_name = "type_name"
        skip_child += 1
        child_txt = self.restructure_value(first_child.text, first_child.tag)

        section_nm_reslvr = f"{elmt_attr[key_type_name]}_{child_txt}"
        section_nm_reslvr = self.check_for_siblings_with_same_name(
            section_nm_reslvr, element_
        )

        parent_path = f"{parent_path}/{section_nm_reslvr}"

        return parent_path, skip_child

    def parse_struct(self, element_: EmtT.Element, parent_path: str) -> None:
        """
        Parameters
        ----------
        element_ : Element with struct tag
        parent_path : Xpath of the parent element where the element_ belongs

        Returns
        -------
        None
        """

        child_num = len(element_)
        elmt_attr = element_.attrib

        # Resolving struct name section is here
        skip_child = -1
        section_nm_reslvr = ""
        first_child = element_[0]
        second_child = element_[1]

        key_name = "name"
        key_value = "value"
        key_type_name = "type_name"
        if key_name in elmt_attr.keys():
            section_nm_reslvr = elmt_attr[key_name]
            section_nm_reslvr = self.check_for_siblings_with_same_name(
                section_nm_reslvr, element_
            )
            parent_path, self.tail_part_frm_struct = self.check_last_part_repetition(
                parent_path, self.tail_part_frm_struct, section_nm_reslvr
            )

        elif key_name not in elmt_attr.keys():
            if (
                key_name in first_child.attrib.values()
                and key_value in second_child.attrib.values()
            ):
                parent_path, skip_child = self.struct_fc_name_sc_value(
                    element_, first_child, parent_path, skip_child
                )
            elif (
                key_name in first_child.attrib.values() and first_child.tag == "string"
            ):
                parent_path, skip_child = self.struct_fc_name_sc_string(
                    element_, first_child, parent_path, skip_child
                )

            else:
                # Check twin siblings
                section_nm_reslvr = self.restructure_value(
                    elmt_attr[key_type_name], "string"
                )
                section_nm_reslvr = (
                    section_nm_reslvr + "_" + str(elmt_attr["__odr_siblings__"])
                )
                parent_path = f"{parent_path}/{section_nm_reslvr}"

        child_elmt_ind = 0
        while child_num > 0:
            self.pass_child_through_parsers(
                element_, parent_path, child_elmt_ind, skip_child
            )
            child_num -= 1
            child_elmt_ind += 1

    def last_element_parser(self, element_: EmtT.Element, parent_path: str) -> None:
        """

        Parameters
        ----------
        element_ : Element with a tag among 'ulong', 'double', 'string',
                   'boolean', 'enum', 'any'
        parent_path : Xpath of the parent element where the element_ belongs

        Returns
        -------
        None
        """

        child_num = len(element_)
        elmt_attr = element_.attrib

        if child_num == 0:
            if "name" in elmt_attr.keys():
                section_nm_reslvr = f'{elmt_attr["name"]}'
                value = self.restructure_value(element_.text, element_.tag)

                parent_path, self.tail_part_frm_othr = self.check_last_part_repetition(
                    parent_path, self.tail_part_frm_othr, section_nm_reslvr
                )
                self.metadata_dict[f"{parent_path}"] = value
            else:
                self.metadata_dict[f"{parent_path}"] = self.restructure_value(
                    element_.text, element_.tag
                )
        elif child_num == 1 and "any" == element_.tag:
            child_elmt = element_[0]
            self.metadata_dict[f"{parent_path}"] = self.restructure_value(
                child_elmt.text, child_elmt.tag
            )

    def check_for_siblings_with_same_name(self, reslv_name, new_sblings_elmt):
        """Check for the same name in the same level. For elments with the same
        write the name _1, _2... .
        """
        child_nm_reslvr_li = new_sblings_elmt.attrib["__parent__"].attrib[
            self.child_nm_reslvers
        ]
        if reslv_name not in child_nm_reslvr_li:
            parent = new_sblings_elmt.attrib["__parent__"]
            parent.attrib[self.child_nm_reslvers].append(reslv_name)
        else:
            last_twin_sib_nm = child_nm_reslvr_li[-1]
            try:
                ind = last_twin_sib_nm.split("_")[-1]
                reslv_name = f"{reslv_name}_{int(ind) + 1}"
                parent = new_sblings_elmt.attrib["__parent__"]
                parent.attrib[self.child_nm_reslvers].append(reslv_name)
            except ValueError:
                reslv_name = f"{reslv_name}_1"
                parent = new_sblings_elmt.attrib["__parent__"]
                parent.attrib[self.child_nm_reslvers].append(reslv_name)
        return reslv_name

    def check_last_part_repetition(
        self, parent_path: str, pre_tail_part: str, new_tail_part: str
    ) -> Tuple[str, str]:
        """
         Check for the data from the same group, for example repetition of the
         experiments under the same physical circumstances, make number of
         them with ..._1, ..._2.
        Parameters
        ----------
        self : XmlSpecs object
        parent_path : Xpath of the parent element
        pre_tail_part : The tail part added in the previous step
        new_tail_part : The tail part obtained from present element

        Returns
        -------
        parent_path : Newly obtained or replaced parent Xpath
        pre_tail_part : Newly obtained or replaced tail_path
        """
        if new_tail_part == pre_tail_part:
            previous_key = f"{parent_path}/{new_tail_part}"
            previous_val = self.metadata_dict.get(previous_key, None)
            if previous_val:
                self.metadata_dict[f"{parent_path}/{new_tail_part}_0"] = previous_val
                pre_tail_part = f"{new_tail_part}_1"
                parent_path = f"{parent_path}/{pre_tail_part}"

                del self.metadata_dict[previous_key]

                return parent_path, pre_tail_part

            parent_path = f"{parent_path}/{pre_tail_part}"
            return parent_path, pre_tail_part

        if new_tail_part in pre_tail_part:
            try:
                ind_ = pre_tail_part.split("_")[-1]
                ind = int(ind_)
                pre_tail_part = f"{new_tail_part}_{ind + 1}"
                parent_path = f"{parent_path}/{pre_tail_part}"
                return parent_path, pre_tail_part
            except ValueError:
                parent_path = f"{parent_path}/{new_tail_part}"
                pre_tail_part = new_tail_part

                return parent_path, pre_tail_part
            except TypeError:
                parent_path = f"{parent_path}/{new_tail_part}"
                pre_tail_part = new_tail_part

                return parent_path, pre_tail_part

        parent_path = f"{parent_path}/{new_tail_part}"
        pre_tail_part = new_tail_part

        return parent_path, pre_tail_part

    @staticmethod
    def restructure_value(value_text: str, element_tag: str) -> Any:
        """
            Collect the value_text transform it to the data_type according
            to the type name provided by element_tag.
        Parameters
        ----------
        value_text : text data that would be 'unsigned long', 'double', 'string',
                    'boolean', 'enum/string'
        element_tag : tag name among 'unsigned long', 'double', 'string',
                    'boolean', 'enum/string'

        Returns
        -------

        """

        def double_(para):
            return np.double(para)

        def ulong_(para):
            return np.uint(para)

        def bool_(para):
            return np.bool_(para)

        data_ty = {"double": double_, "ulong": ulong_, "boolean": bool_}
        string_ty = ["string", "enum"]

        if not value_text:
            return ""

        if element_tag in string_ty:
            value_text_: Any = " ".join(value_text.split()).replace(" ", "_")
            return value_text_

        for key_, _ in data_ty.items():
            if key_ == element_tag:
                value_text_ = value_text.split()
                numpy_value = data_ty[element_tag](value_text_)[...]
                if np.shape(numpy_value) == (1,):
                    return numpy_value[0]
                return numpy_value

    def cumulate_counts_series(
        self,
        scan_seq_elem: EmtT.Element,
        counts_length: int = None,
        cumulative_counts: np.ndarray = None,
    ) -> Tuple[str, np.ndarray]:
        """
        Sum the counts over different scans. Each ScanSeaq contains
        multiple scans under the same physical environment. The
        multiple scans are usually taken to make the peaks visible and
        distinguishable.

        Parameters
        ----------
        scan_seq_elem : Element with ScanSeq tag
        counts_length : Number of count (length of 1D numpy array)
                        contain in each scan
        cumulative_counts : Cumulative counts up to last scan  from the
                            same ScanSeq

        Returns
        -------
        np.ndarray : Cumulative up to last scans from the same ScanSeq
        """

        child_num = len(scan_seq_elem)
        name = "count"

        child_elmt_ind = 0
        while child_num >= 0:
            if scan_seq_elem.attrib["type_name"] == "CountsSeq":
                num_of_counts = int(scan_seq_elem.attrib["length"])
                if not counts_length:
                    counts_length = num_of_counts
                if counts_length != num_of_counts:
                    raise ValueError("Count number from all scans must be equal!!")

            if scan_seq_elem.attrib["type_name"] == "Counts":
                counts_data = self.restructure_value(
                    scan_seq_elem.text, scan_seq_elem.tag
                )

                if cumulative_counts is None:
                    cumulative_counts = counts_data
                else:
                    cumulative_counts = cumulative_counts + counts_data

            if child_num > 0:
                child_element = scan_seq_elem[child_elmt_ind]
                name, cumulative_counts = self.cumulate_counts_series(
                    child_element, counts_length, cumulative_counts
                )

            child_num = child_num - 1
            child_elmt_ind = child_elmt_ind + 1

        return (name, cumulative_counts)

    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    def collect_raw_data_to_construct_data(self):
        """Collect the raw data about detectors so that the binding energy and
        and counts for the corresponding nominal.
        """

        entry_list: List = []
        raw_dict = {
            "mcd_num": 0,
            "curves_per_scan": 0,
            "values_per_curve": 0,
            "mcd_head": 0,
            "mcd_tail": 0,
            "excitation_energy": 0,
            "kinetic_energy": 0,
            "kinetic_energy_base": 0,
            "effective_workfunction": 0,
            "scan_delta": 0,
            "pass_energy": 0,
            "mcd_shifts": [],
            "mcd_poss": [],
            "mcd_gains": [],
            "time": 0,
            "scans": {},
        }
        for key, val in self.metadata_dict.items():
            entry = construct_entry_name(key)

            if entry and (entry not in entry_list):
                self.entry_to_data[entry] = {"raw_data": copy.deepcopy(raw_dict)}
                entry_list.append(entry)

            if "region/curves_per_scan" in key:
                self.entry_to_data[entry]["raw_data"]["curves_per_scan"] = val
            elif "region/values_per_curve" in key:
                self.entry_to_data[entry]["raw_data"]["values_per_curve"] = val

            elif "region/excitation_energy" in key:
                self.entry_to_data[entry]["raw_data"]["excitation_energy"] = val

            elif "region/scan_mode/name" in key:
                val = self._convert_energy_scan_mode(val)
                self.entry_to_data[entry]["raw_data"]["scan_mode"] = val

            elif "region/kinetic_energy" in key:
                if "region/kinetic_energy_base" not in key:
                    self.entry_to_data[entry]["raw_data"]["kinetic_energy"] = val
                    continue
                if "region/kinetic_energy_base" in key:
                    self.entry_to_data[entry]["raw_data"]["kinetic_energy_base"] = val
                    continue

            elif "region/effective_workfunction" in key:
                self.entry_to_data[entry]["raw_data"]["effective_workfunction"] = val

            elif "region/scan_delta" in key:
                self.entry_to_data[entry]["raw_data"]["scan_delta"] = val

            elif "region/pass_energy" in key:
                self.entry_to_data[entry]["raw_data"]["pass_energy"] = val

            elif "mcd_head" in key:
                self.entry_to_data[entry]["raw_data"]["mcd_head"] = val

            elif "mcd_tail" in key:
                self.entry_to_data[entry]["raw_data"]["mcd_tail"] = val

            elif "shift" in key:
                self.entry_to_data[entry]["raw_data"]["mcd_shifts"].append(val)
                self.entry_to_data[entry]["raw_data"]["mcd_num"] += 1

            elif "gain" in key:
                self.entry_to_data[entry]["raw_data"]["mcd_gains"].append(val)

            elif "position" in key:
                self.entry_to_data[entry]["raw_data"]["mcd_poss"].append(val)

            # construct scan names e.g cycles2_scan0
            if "cycles/Cycle_" in key:
                _, last_part = key.split("cycles/Cycle_")
                if "/time" in last_part:
                    self.entry_to_data[entry]["raw_data"]["time"] = val
                    continue
                if "/parameters/Loop" in last_part:
                    self.entry_to_data[entry]["raw_data"]["loop_no"] = val
                    continue
                parts = last_part.split("/")
                cycle_num, scan_num = parts[0], parts[-2].split("_")[1]
                scan_name = f"cycle{cycle_num}_scan{scan_num}"

                self.entry_to_data[entry]["raw_data"]["scans"][scan_name] = val

    def _convert_energy_scan_mode(self, energy_scan_mode):
        """
        Convert the native names for the energy scan modes to the ones
        used in NXmpes.

        """
        energy_scan_mode_map = {
            "FixedAnalyzerTransmission": "fixed_analyser_transmission",
            "FixedRetardationRatio": "fixed_retardation_ratio",
            "FixedEnergies": "fixed_energy",
            "Snapshot": "snapshot",
        }

        try:
            energy_scan_mode = energy_scan_mode_map[energy_scan_mode]
        except KeyError:
            pass
        return energy_scan_mode



"""
Classes for reading XPS files from TXT export of CasaXPS.
"""
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=too-many-lines,too-few-public-methods

import itertools
import operator
import copy
from abc import ABC, abstractmethod
import xarray as xr
import numpy as np

from pynxtools_xps.reader_utils import (
    XPSMapper,
    check_uniform_step_width,
    get_minimal_step,
    interpolate_arrays,
    construct_entry_name,
    construct_data_key,
    construct_detector_data_key,
)


class TxtMapperVamasExport(XPSMapper):
    """
    Class for restructuring .txt data file from
    Casa TXT export (from Vamas) into python dictionary.
    """

    config_file = "config_txt_vamas_export.json"

    def __init__(self):
        self.parser_map = {
            "rows_of_tables": TextParserRows,
            "columns_of_tables": TextParserColumns,
        }
        super().__init__()

    def _get_file_type(self, file):
        """
        Check which export option was used in CasaXPS.

        Parameters
        ----------
        file : str
            XPS data filepath.

        Returns
        -------
        str
            Either columns_of_tables or rows_of_tables.

        """
        with open(file, encoding="utf-8") as txt_file:
            first_line = txt_file.readline()
            if first_line.startswith("Cycle"):
                return "columns_of_tables"
            return "rows_of_tables"

    def _select_parser(self):
        """
        Select parser based on the structure of the text file

        Returns
        -------
        TextParser
            TextParser for CasaXPS export from Vamas files.

        """
        return self.parser_map[self._get_file_type(self.file)]()

    def construct_data(self):
        """Map TXT format to NXmpes-ready dict."""
        # pylint: disable=duplicate-code
        spectra = copy.deepcopy(self.raw_data)

        self._xps_dict["data"]: dict = {}

        key_map = {
            "file_info": [],
            "user": [],
            "instrument": [],
            "source": [],
            "beam": [
                "excitation_energy",
                "excitation_energy/units",
            ],
            "analyser": [],
            "collectioncolumn": [],
            "energydispersion": [],
            "detector": [
                "dwell_time",
                "dwell_time/units",
            ],
            "manipulator": [],
            "calibration": [],
            "sample": [],
            "data": [
                "dwell_time",
                "x_units" "y_units",
                "start_energy",
                "stop_energy",
                "step_size",
            ],
            "region": ["region_name"],
        }

        for spectrum in spectra:
            self._update_xps_dict_with_spectrum(spectrum, key_map)

    def _update_xps_dict_with_spectrum(self, spectrum, key_map):
        """
        Map one spectrum from raw data to NXmpes-ready dict.

        """
        # pylint: disable=too-many-locals,duplicate-code
        group_parent = f'{self._root_path}/Group_{spectrum["group_name"]}'
        region_parent = f'{group_parent}/regions/Region_{spectrum["spectrum_type"]}'
        file_parent = f"{region_parent}/file_info"
        instrument_parent = f"{region_parent}/instrument"
        analyser_parent = f"{instrument_parent}/analyser"

        path_map = {
            "file_info": f"{file_parent}",
            "user": f"{region_parent}/user",
            "instrument": f"{instrument_parent}",
            "source": f"{instrument_parent}/source",
            "beam": f"{instrument_parent}/beam",
            "analyser": f"{analyser_parent}",
            "collectioncolumn": f"{analyser_parent}/collectioncolumn",
            "energydispersion": f"{analyser_parent}/energydispersion",
            "detector": f"{analyser_parent}/detector",
            "manipulator": f"{instrument_parent}/manipulator",
            "calibration": f"{instrument_parent}/calibration",
            "sample": f"{region_parent}/sample",
            "data": f"{region_parent}/data",
            "region": f"{region_parent}",
        }

        for grouping, spectrum_keys in key_map.items():
            root = path_map[str(grouping)]
            for spectrum_key in spectrum_keys:
                try:
                    mpes_key = spectrum_key
                    self._xps_dict[f"{root}/{mpes_key}"] = spectrum[spectrum_key]
                except KeyError:
                    pass

        # Create keys for writing to data and detector
        entry = construct_entry_name(region_parent)
        scan_key = construct_data_key(spectrum)
        detector_data_key_child = construct_detector_data_key(spectrum)
        detector_data_key = f'{path_map["detector"]}/{detector_data_key_child}/counts'

        energy = np.array(spectrum["data"]["binding_energy"])
        intensity = np.array(spectrum["data"]["intensity"])

        # If multiple spectra exist to entry, only create a new
        # xr.Dataset if the entry occurs for the first time.
        if entry not in self._xps_dict["data"]:
            self._xps_dict["data"][entry] = xr.Dataset()

        # Write averaged cycle data to 'data'.
        all_scan_data = [
            value
            for key, value in self._xps_dict["data"][entry].items()
            if scan_key.split("_")[0] in key
        ]
        averaged_scans = np.mean(all_scan_data, axis=0)
        if averaged_scans.size == 1:
            # on first scan in cycle
            averaged_scans = intensity

        self._xps_dict["data"][entry][scan_key.split("_")[0]] = xr.DataArray(
            data=averaged_scans,
            coords={"energy": energy},
        )

        self._xps_dict["data"][entry][scan_key] = xr.DataArray(
            data=intensity, coords={"energy": energy}
        )

        self._xps_dict[detector_data_key] = intensity


class TextParser(ABC):  # pylint: disable=too-few-public-methods
    """
    Parser for ASCI files exported from CasaXPS.
    """

    def __init__(self):
        self.lines = []
        self.data_dict = []
        self.n_headerlines = 7
        self.uniform_energy_steps = True

    def parse_file(self, file, uniform_energy_steps=True, **kwargs):
        """
        Parse the file into a list of dictionaries.

        Parsed data stored in the attribute 'self.data'.

        Parameters
        ----------
        file : str
            XPS data filepath.
        uniform_energy_steps : bool, optional
            If true, the spectra are interpolate to have uniform
            energy steps. The default is True.
        **kwargs : dict
            n_headerlines: number of header_lines in each data block.

        Returns
        -------
        dict
            DESCRIPTION.

        """
        if "n_headerlines" in kwargs:
            self.n_headerlines = kwargs["n_headerlines"]

        self.uniform_energy_steps = uniform_energy_steps

        self._read_lines(file)
        blocks = self._parse_blocks()
        return self._build_list_of_dicts(blocks)

    def _read_lines(self, file):
        """
        Read in all lines from the file as a list of strings.

        Parameters
        ----------
        file : str
            XPS data filepath.

        Returns
        -------
        None.

        """
        with open(file, encoding="utf-8") as txt_file:
            for line in txt_file:
                self.lines += [line]

    @abstractmethod
    def _parse_blocks(self):
        """
        Extract spectrum blocks from full data string.

        This method has to be implemented in the inherited parsers.

        Returns
        -------
        blocks : list
            List of strings, with each string containing one spectrum's
            data and metadata.

        """
        blocks = []

        return blocks

    @abstractmethod
    def _build_list_of_dicts(self, blocks):
        """
        Build list of dictionaries, with each dict containing data
        and metadata of one spectrum (block).

        This method has to be implemented in the inherited parsers.

        Parameters
        ----------
        blocks : list
            List of data blocks containing one spectrum each.

        Returns
        -------
        spectra : list
            List of dicts with spectrum data and metadata.

        """
        spectra = []

        return spectra

    def _separate_header_and_data(self, block):
        """
        Separate header (with metadata) from data for one measurement
        block

        Returns
        -------
        None.

        """
        header = block[: self.n_headerlines]
        data = block[self.n_headerlines :]

        return header, data


class TextParserRows(TextParser):
    """
    Parser for ASCI files exported from CasaXPS using the
    'Rows of Tables' option.
    """

    def __init__(self):
        super().__init__()
        self.n_headerlines = 7

    def _parse_blocks(self):
        """
        With the 'Rows of Tables' option, there is only one block
        with common metadata.

        """
        return self.lines

    def _build_list_of_dicts(self, blocks):
        """
        Build list of dictionaries, with each dict containing data
        and metadata of one spectrum.

        Parameters
        ----------
        blocks : list
            List of data blocks containing one spectrum each.

        Returns
        -------
        spectra : list
            List of dicts with spectrum data and metadata.

        """
        spectra = []

        header, data_lines = self._separate_header_and_data(blocks)
        settings = self._parse_header(header)
        data = self._parse_data(data_lines)
        for spec_settings, spec_data in zip(settings, data):
            spectra += [{**spec_settings, **spec_data}]

        return spectra

    def _parse_header(self, header):
        """
        Parse header into metadata dictionary.

        Parameters
        ----------
        header : str
            Header data for one spectrum as a String.

        Returns
        -------
        settings : list
            List of dicts with measurement settings for
            one spectrum each.

        """
        settings = []
        for spec_header in header[-1].split("\t")[1::3]:
            group_name = spec_header.split(":")[1]
            region = spec_header.split(":")[2]
            spectrum_settings = {
                "group_name": group_name,
                "spectrum_type": region,
                "x_units": "binding",
                "y_units": spec_header.split(":")[3],
            }
            settings += [spectrum_settings]

        return settings

    def _parse_data(self, data_lines):
        """
        Extract energy and intensity data.

        Parameters
        ----------
        data_lines : list
            List of lines with measurement data.

        Returns
        -------
        list
            List of dicts containing the binding energy
            and the intensity axes of one spectrum each.

        """
        data_lines = [x.split("\t") for x in data_lines]
        for line in data_lines:
            del line[2::3]
            del line[-1]

        lines = [[] for _ in range(max(len(line) for line in data_lines))]

        for line in data_lines:
            for i, data_point in enumerate(line):
                try:
                    lines[i].append(float(data_point.strip("\n")))
                except ValueError:
                    pass

        data = []

        for x_bin, intensity in zip(lines[::2], lines[1::2]):
            x_bin, intensity = np.array(x_bin), np.array(intensity)

            if self.uniform_energy_steps and not check_uniform_step_width(x_bin):
                x_bin, intensity = interpolate_arrays(x_bin, intensity)

            spectrum = {
                "data": {
                    "binding_energy": np.array(x_bin),
                    "intensity": np.array(intensity).squeeze(),
                },
                "start_energy": x_bin[0],
                "stop_energy": x_bin[-1],
                "x_units": "binding",
                "y_units": "CPS",
            }

            if check_uniform_step_width(x_bin):
                spectrum["step_size"] = get_minimal_step(x_bin)

            data += [spectrum]

        return data


class TextParserColumns(TextParser):
    """
    Parser for ASCI files exported from CasaXPS using the
    'Columns of Tables' option.
    """

    def __init__(self):
        super().__init__()
        self.n_headerlines = 8

    def _parse_blocks(self):
        """
        Extract spectrum blocks from full data string.

        Returns
        -------
        blocks : list
            List of strings, with each string containing one spectrum's
            data and metadata.

        """
        blocks = [
            list(g) for _, g in itertools.groupby(self.lines, lambda i: "Cycle " in i)
        ]
        blocks = [operator.add(*blocks[i : i + 2]) for i in range(0, len(blocks), 2)]

        return blocks

    def _parse_block_header(self, header):
        """
        Parse spectrum block header into metadata dictionary.

        Parameters
        ----------
        header : str
            Header data for one spectrum as a String.

        Returns
        -------
        settings : dictf
            Dict of measurement settings for one spectrum.

        """
        group_name = header[0].split(":")[1]
        region = header[0].split(":")[2].split("\n")[0]
        settings = {
            "group_name": group_name,
            "spectrum_type": region,
            "excitation_energy": header[1].split("\t")[2],
            "excitation_energy/@units": header[1].split("\t")[1].split(" ")[-1],
            "dwell_time": float(header[1].split("\t")[4].strip()),
            "dwell_time/@units": header[1].split("\t")[3].split(" ")[-1],
        }

        return settings

    def _parse_block_data(self, block_data):
        """
        Extract energy and intensity data from one data block.

        Parameters
        ----------
        block_data : list
            List of lines with measurement data.

        Returns
        -------
        dict
            Dict containing the kinetic/binding energy
            and the intensity axes.

        """
        lines = np.array([[float(i) for i in d.split()] for d in block_data])

        x_kin = lines[:, 0]
        x_bin = lines[:, 2]
        intensity = lines[:, 1]

        if self.uniform_energy_steps and not check_uniform_step_width(x_kin):
            x_kin, (x_bin, intensity) = interpolate_arrays(x_kin, [x_bin, intensity])

        return {
            "kinetic_energy": np.array(x_kin),
            "binding_energy": np.array(x_bin),
            "intensity": np.array(intensity).squeeze(),
        }

    def _build_list_of_dicts(self, blocks):
        """
        Build list of dictionaries, with each dict containing data
        and metadata of one spectrum (block).


        Parameters
        ----------
        blocks : list
            List of data blocks containing one spectrum each.

        Returns
        -------
        spectra : list
            List of dicts with spectrum data and metadata.

        """
        spectra = []
        for block in blocks:
            header, block_data_lines = self._separate_header_and_data(block)
            block_settings = self._parse_block_header(header)
            block_data = {"data": self._parse_block_data(block_data_lines)}
            kinetic_energy = block_data["data"]["kinetic_energy"]
            block_settings.update(
                {
                    "start_energy": kinetic_energy[0],
                    "stop_energy": kinetic_energy[-1],
                }
            )
            if check_uniform_step_width(kinetic_energy):
                block_settings["step_size"] = get_minimal_step(kinetic_energy)

            spectra += [{**block_settings, **block_data}]

        return spectra



"""
A generic reader for loading XPS (X-ray Photoelectron Spectroscopy) data
file into mpes nxdl (NeXus Definition Language) template.
"""
import datetime
import sys

# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path
from typing import Any, Dict, List, Set, Tuple

import numpy as np
import yaml

from pynxtools.dataconverter.helpers import extract_atom_types
from pynxtools.dataconverter.readers.base.reader import BaseReader
from pynxtools.dataconverter.readers.utils import (
    FlattenSettings,
    flatten_and_replace,
    parse_flatten_json,
)
from pynxtools.dataconverter.template import Template
from pynxtools_xps.file_parser import XpsDataFileParser
from pynxtools_xps.reader_utils import construct_entry_name

np.set_printoptions(threshold=sys.maxsize)

XPS_TOKEN = "@xps_token:"
XPS_DATA_TOKEN = "@data:"
XPS_DETECTOR_TOKEN = "@detector_data:"
ELN_TOKEN = "@eln"
LINK_TOKEN = "link"
TOKEN_SET = {XPS_TOKEN, XPS_DATA_TOKEN, XPS_DETECTOR_TOKEN, ELN_TOKEN}

# Track entries for using for eln data
ENTRY_SET: Set[str] = set()
DETECTOR_SET: Set[str] = set()
POSSIBLE_ENTRY_PATH: Dict = {}


CONVERT_DICT = {
    "unit": "@units",
    "version": "@version",
    "user": "USER[user]",
    "instrument": "INSTRUMENT[instrument]",
    "source_probe": "source_TYPE[source_probe]",
    "beam_probe": "beam_TYPE[beam_probe]",
    "analyser": "ELECTRONANALYSER[electronanalyser]",
    "collectioncolumn": "COLLECTIONCOLUMN[collectioncolumn]",
    "energydispersion": "ENERGYDISPERSION[energydispersion]",
    "detector": "DETECTOR[detector]",
    "manipulator": "MANIPULATOR[manipulator]",
    "pid": "PID[pid]",
    "process": "PROCESS[process]",
    "sample": "SAMPLE[sample]",
    "substance": "SUBSTANCE[substance]",
    # "Data": "DATA[data]",
}

REPLACE_NESTED: Dict[str, str] = {}


def find_entry_and_value(xps_data_dict, key_part, dt_typ):
    """Construct the entry name and pick up the corresponding data for
    that for that entry.
    """

    entries_values = {}
    if dt_typ == XPS_TOKEN:
        for key, val in xps_data_dict.items():
            if key.endswith(key_part):
                entry = construct_entry_name(key)
                entries_values[entry] = val

    elif dt_typ in (XPS_DATA_TOKEN, XPS_DETECTOR_TOKEN):
        # entries_values = entry:{cycle0_scan0_chan0:xr.data}
        entries_values = xps_data_dict["data"]

    return entries_values


def get_entries_and_detectors(config_dict, xps_data_dict):
    """Get all entries in the xps_data_dict"""
    for key, value in config_dict.items():
        for token in [XPS_DATA_TOKEN, XPS_DETECTOR_TOKEN, XPS_TOKEN]:
            try:
                key_part = value.split(token)[-1]
                entries_values = find_entry_and_value(
                    xps_data_dict, key_part, dt_typ=token
                )

                for entry, data in entries_values.items():
                    if entry:
                        ENTRY_SET.add(entry)
                    if token == XPS_DETECTOR_TOKEN:
                        chan_count = "_chan"
                        # Iteration over scan
                        for data_var in data.data_vars:
                            if chan_count in data_var:
                                detector_num = data_var.split(chan_count)[-1]
                                detector_nm = f"detector{detector_num}"
                                DETECTOR_SET.add(detector_nm)
            except AttributeError:
                continue


# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
def fill_data_group(key, key_part, entries_values, config_dict, template):
    """Fill out fields and attributes for NXdata"""

    survey_count_ = 0
    count = 0

    for entry, xr_data in entries_values.items():
        root = key[0]

        # Set first Survey as default for .nxs file
        if "Survey" in entry and survey_count_ == 0:
            survey_count_ += 1
            template[f"{root}@default"] = entry

        # If no Survey set any scan for default
        if survey_count_ == 0 and count == 0:
            count += 1
            template[f"{root}@default"] = entry

        data_field_key = key.replace("entry", entry)
        data_group_key = data_field_key.rsplit("/data", 1)[0]

        if key_part == "energy":
            # Here, we need to use the energy value if the data/energy is
            # written directly in the future.
            energy_field_key = data_group_key
            energy = np.array(xr_data.coords["energy"].values)
            template[energy_field_key] = energy

        else:
            # Define energy axis and energy_indices
            template[f"{data_group_key}/energy/@long_name"] = "energy"
            template[f"{data_group_key}/@energy_indices"] = 0

            units = "counts_per_second"

            chan_count = "_chan"
            scan_count = "_scan"
            for data_var in xr_data.data_vars:
                cycle_scan = data_var
                # Collecting only accumulated counts
                # indivisual channeltron counts goes to detector data section
                if chan_count not in cycle_scan and scan_count in cycle_scan:
                    indv_scan_key = f"{data_group_key}/{cycle_scan}"
                    indv_scan_key_unit = f"{indv_scan_key}/@units"
                    template[indv_scan_key] = xr_data[data_var].data
                    template[indv_scan_key_unit] = units

            template[data_field_key] = np.mean(
                [
                    xr_data[x_arr].data
                    for x_arr in xr_data.data_vars
                    if (chan_count not in x_arr and scan_count in x_arr)
                ],
                axis=0,
            )

            template[f"{data_field_key}_errors"] = np.std(
                [
                    xr_data[x_arr].data
                    for x_arr in xr_data.data_vars
                    if (chan_count not in x_arr and scan_count in x_arr)
                ],
                axis=0,
            )

            template[f"{data_field_key}/@long_name"] = "XPS intensity"


def fill_detector_group(key, entries_values, config_dict, xps_data_dict, template):
    """Fill out fileds and attributes for NXdetector/NXdata"""

    for entry, xr_data in entries_values.items():
        chan_count = "_chan"

        # unit_key = config_dict[f"{key}/@units"]
        # units = find_entry_and_value(xps_data_dict, unit_key, XPS_TOKEN)
        units = "counts_per_second"

        # Iteration over scan
        data_vars = [
            data_var for data_var in xr_data.data_vars if chan_count in data_var
        ]

        detector_scans = {detector: [] for detector in DETECTOR_SET}

        for data_var in data_vars:
            if chan_count in data_var:
                detector_num = data_var.split(chan_count)[-1]
                detector_nm = f"detector{detector_num}"
                detector_scans[detector_nm] += [xr_data[data_var].data]
                cycle_scan_num = data_var.split(chan_count)[0]
                modified_key = key.replace("entry", entry)
                modified_key = modified_key.replace("[detector]", f"[{detector_nm}]")
                cycle_scan_key = modified_key.replace(
                    "raw_data/raw", f"raw_data/{cycle_scan_num}"
                )
                template[cycle_scan_key] = xr_data[data_var].data
                template[f"{cycle_scan_key}/@units"] = units

        # Add multi-dimensional `raw`array for each detector
        for detector_nm, value in detector_scans.items():
            modified_key = key.replace("entry", entry)
            modified_key = modified_key.replace("[detector]", f"[{detector_nm}]")
            template[modified_key] = np.array(value)


def fill_template_with_value(key, value, template):
    """
    Fill NeXus template with a key-value pair.

    Parameters
    ----------
    key : str
        DESCRIPTION.
    value :
        Any value coming from the XPS, config, or ELN file.
    template : Template
        A NeXus template.

    """
    if value is None or str(value) == "None":
        return

    atom_types: List = []
    if "chemical_formula" in key:
        atom_types = list(extract_atom_types(value))

    if isinstance(value, datetime.datetime):
        value = value.isoformat()

    elif isinstance(value, dict) and LINK_TOKEN in value:
        initial_link_text = value[LINK_TOKEN]

    # Do for all entry names
    for entry in ENTRY_SET:
        if isinstance(value, dict) and LINK_TOKEN in value:
            # Reset link to original
            value[LINK_TOKEN] = initial_link_text.replace("entry", f"{entry}")

        modified_key = key.replace("[entry]", f"[{entry}]")

        # Do for all detectors
        if "[detector]" in key:
            for detector in DETECTOR_SET:
                detr_key = modified_key.replace("[detector]", f"[{detector}]")
                template[detr_key] = value

                if isinstance(value, dict) and LINK_TOKEN in value:
                    link_text = value[LINK_TOKEN]
                    if "/detector/" in link_text:
                        # Only replace if generic detector is given in
                        # link.
                        link_text = link_text.replace("detector", f"{detector}")
                        value = {LINK_TOKEN: link_text}
                template[detr_key] = value

        else:
            template[modified_key] = value

        if atom_types:
            modified_key = modified_key.replace("chemical_formula", "atom_types")
            template[modified_key] = ", ".join(atom_types)


def fill_template_with_xps_data(config_dict, xps_data_dict, template):
    """Collect the xps data from xps_data_dict
    and store them into template. We use searching_keys
    for separating the data from xps_data_dict.
    """
    for key, config_value in config_dict.items():
        if isinstance(config_value, str) and any(
            token in config_value for token in TOKEN_SET
        ):
            if XPS_DATA_TOKEN in str(config_value):
                key_part = config_value.split(XPS_DATA_TOKEN)[-1]
                entries_values = find_entry_and_value(
                    xps_data_dict, key_part, dt_typ=XPS_DATA_TOKEN
                )
                fill_data_group(key, key_part, entries_values, config_dict, template)

            elif XPS_DETECTOR_TOKEN in str(config_value):
                key_part = config_value.split(XPS_DATA_TOKEN)[-1]
                entries_values = find_entry_and_value(
                    xps_data_dict, key_part, dt_typ=XPS_DETECTOR_TOKEN
                )

                fill_detector_group(
                    key, entries_values, config_dict, xps_data_dict, template
                )

            elif XPS_TOKEN in str(config_value):
                key_part = config_value.split(XPS_TOKEN)[-1]
                entries_values = find_entry_and_value(
                    xps_data_dict, key_part, dt_typ=XPS_TOKEN
                )
                for entry, ent_value in entries_values.items():
                    modified_key = key.replace("[entry]", f"[{entry}]")
                    fill_template_with_value(modified_key, ent_value, template)

        else:
            fill_template_with_value(key, config_value, template)


def fill_template_with_eln_data(eln_data_dict, config_dict, template):
    """Fill the template from provided eln data"""
    for key, config_value in config_dict.items():
        if ELN_TOKEN in str(config_value):
            try:
                field_value = eln_data_dict[key]
                fill_template_with_value(key, field_value, template)
            except KeyError:
                pass
        elif key in eln_data_dict:
            field_value = eln_data_dict[key]
            fill_template_with_value(key, field_value, template)


def concatenate_values(value1, value2):
    """
    Concatenate two values of same type to be stored
    in xps_data_dict. Dicts are merged and every other object is
    appended to a list.

    """
    if isinstance(value1, dict) and isinstance(value2, dict):
        concatenated = {**value1, **value2}
    else:
        if not isinstance(value1, list):
            value1 = [value1]
        if not isinstance(value2, list):
            value2 = [value2]
        concatenated = value1 + value2

    return concatenated


# pylint: disable=too-few-public-methods
class XPSReader(BaseReader):
    """Reader for XPS."""

    supported_nxdls = [
        "NXmpes",
    ]

    def read(
        self,
        template: dict = None,
        file_paths: Tuple[str] = None,
        objects: Tuple[Any] = None,
        **kwargs,
    ) -> dict:
        """Reads data from given file and returns
        a filled template dictionary"""

        reader_dir = Path(__file__).parent

        xps_data_dict: Dict[str, Any] = {}
        eln_data_dict: Dict[str, Any] = {}
        config_file: Path = reader_dir.joinpath("config", "template.json")

        for file in file_paths:
            file_ext = file.rsplit(".")[-1]

            if file_ext in ["yaml", "yml"]:
                with open(file, mode="r", encoding="utf-8") as eln:
                    eln_data_dict = flatten_and_replace(
                        FlattenSettings(
                            yaml.safe_load(eln), CONVERT_DICT, REPLACE_NESTED
                        )
                    )
            elif file_ext in XpsDataFileParser.__prmt_file_ext__:
                parser = XpsDataFileParser([file])
                data_dict = parser.get_dict(**kwargs)
                config_file = parser.config_file

                # If there are multiple input data files of the same type,
                # make sure that existing keys are not overwritten.
                existing = [
                    (key, xps_data_dict[key], data_dict[key])
                    for key in set(xps_data_dict).intersection(data_dict)
                ]

                xps_data_dict = {**xps_data_dict, **data_dict}
                for key, value1, value2 in existing:
                    xps_data_dict[key] = concatenate_values(value1, value2)

                config_file = reader_dir.joinpath(f"config/{config_file}")
            elif file_ext in XpsDataFileParser.__prmt_metadata_file_ext__:
                data_dict = XpsDataFileParser([file]).get_dict(**kwargs)

                xps_data_dict = {**xps_data_dict, **data_dict}

            # This code is not very robust.
            elif file_ext == "json":
                if "config" in file:
                    config_file = Path(file)

        config_dict = parse_flatten_json(config_file)

        get_entries_and_detectors(config_dict, xps_data_dict)
        fill_template_with_xps_data(config_dict, xps_data_dict, template)

        if eln_data_dict:
            # Filling in ELN metadata and overwriting the common
            # paths by giving preference to the ELN metadata
            fill_template_with_eln_data(eln_data_dict, config_dict, template)
        else:
            raise ValueError(
                "Eln file must be submited with some required fields and attributes."
            )

        final_template = Template()
        for key, val in template.items():
            if ("/ENTRY[entry]" not in key) and (val is not None):
                final_template[key] = val

        return final_template


READER = XPSReader
